from enum import Enum

from kntgen.file_helpers import ModifiedFileState


class BlockChar:
    TOP_LEFT = '╔'
    TOP_RIGHT = '╗'
    BOTTOM_LEFT = '╚'
    BOTTOM_RIGHT = '╝'
    HORIZON = '═'
    VERTICAL = '║'
    CENTER_LEFT = '╟'
    CENTER_RIGHT = '╢'


class ConsoleColor:
    LIGHT = '\033[95m'
    BLUE = '\033[94m'
    CYAN = '\033[96m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    WHITE = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class PrintMode(Enum):
    INFO = 0
    WARNING = 1
    DEBUG = 2
    ERROR = 3
    NORMAL = 5


def info_printer(content: str, emphasis=False):
    printer(content, mode=PrintMode.INFO, emphasis=emphasis)


def warning_printer(content: str, emphasis=True):
    printer(content, mode=PrintMode.WARNING, emphasis=emphasis)


def debug_printer(content: str, emphasis=False):
    printer(content, mode=PrintMode.DEBUG, emphasis=emphasis)


def error_printer(content: str, emphasis=False):
    printer(content, mode=PrintMode.ERROR, emphasis=emphasis)


def printer(content: str,
            mode: PrintMode = PrintMode.NORMAL,
            emphasis=True):
    color_of_mode = {
        PrintMode.INFO: ConsoleColor.BLUE,
        PrintMode.WARNING: ConsoleColor.YELLOW,
        PrintMode.DEBUG: ConsoleColor.GREEN,
        PrintMode.ERROR: ConsoleColor.RED,
    }.get(mode, ConsoleColor.WHITE)
    console_content = content.upper() if emphasis else content
    print(f'{color_of_mode}{console_content}{ConsoleColor.WHITE}')


def beauty_printer(title: str,
                   content: list[str] = None,
                   content_has_block=True,
                   need_line_separate=True,
                   max_char_inline=100):
    smooth_line = BlockChar.HORIZON * max_char_inline
    dash_line = '╌' * max_char_inline
    beauty_title = f'{ConsoleColor.YELLOW}{title}{ConsoleColor.WHITE}'
    title_color_chars_len = len(f'{ConsoleColor.YELLOW}{ConsoleColor.WHITE}')
    print(f'{BlockChar.TOP_LEFT}{smooth_line}{BlockChar.TOP_RIGHT}')
    print(
        f'{BlockChar.VERTICAL}{beauty_title:^{max_char_inline + title_color_chars_len}}{BlockChar.VERTICAL}')
    print(f'{BlockChar.CENTER_LEFT}{smooth_line}{BlockChar.CENTER_RIGHT}')
    for index, chunk in enumerate(content):
        is_last_chunk = index == len(content) - 1
        no_change = ModifiedFileState.NO_CHANGE in chunk
        has_appended = ModifiedFileState.HAS_APPENDED in chunk
        chunk_color = ConsoleColor.WHITE if no_change else (
            ConsoleColor.BLUE if has_appended else ConsoleColor.GREEN)
        if content_has_block:
            def _print_middle_line(st):
                present_st = ' ' if not st or st.isspace() else st
                print(
                    f'{BlockChar.VERTICAL}{chunk_color}{present_st:{max_char_inline}}{ConsoleColor.WHITE}{BlockChar.VERTICAL}')

            if len(chunk) < max_char_inline or '\033[38;2' in chunk:
                _print_middle_line(chunk)
            else:
                smaller_chunks = []
                lines = chunk.split('\n')
                for line in lines:
                    smaller_chunks.extend(
                        [line[i:i + max_char_inline - 20] for i in
                         range(0, len(line),
                               max_char_inline - 20)])
                for smaller_chunk in smaller_chunks:
                    _print_middle_line(smaller_chunk)
        else:
            print(chunk)
        last_line = f'{BlockChar.BOTTOM_LEFT}{smooth_line}{BlockChar.BOTTOM_RIGHT}' \
            if is_last_chunk \
            else f'{BlockChar.CENTER_LEFT}{dash_line}{BlockChar.CENTER_RIGHT}'
        if need_line_separate or is_last_chunk:
            print(last_line)
    print(ConsoleColor.WHITE)
    print('')


def colored_printer(r, g, b, text):
    return f'\033[38;2;{r:03};{g:03};{b:03}m{text}\033[38;2;255;255;255m'



class UICheatSheet:
    CONCISE_INFO = {
        'wsb': 'White Status Bar',
        'bsb': 'Black Status Bar',
        'sf': 'Scaffold',
        'sfa': 'Scaffold\'s appbar',
        'sfb': 'Scaffold\'s body',
        'ct': 'Container',
        'sb': 'SizedBox',
        'c': 'Column',
        'r': 'Row',
        'w': 'Wrap',
        's': 'Stack',
        'sv': 'Single Child Scroll View',
        'lv': 'List View',
        'lvb': 'List View builder',
        'gv': 'Grid View',
        'gvb': 'Grid View builder',
        'pw': 'Paging Widget',
        'i': 'Image View',
        't': 'Text View',
        'tf': 'Text Field',
        'b': 'Button',
        'bb': 'Bloc Builder',
        'bl': 'Bloc Listener',
        'bp': 'Bloc Provider',
        'mbl': 'Multi Bloc Listener',
        'mbp': 'Multi Bloc Provider',
        'stf': 'Stateful Widget, default is Stateless for custom Widget',
        'dt': 'Default Toolbar',
        'da': 'Default Appbar',
        'wp': 'Custom Web Page',
        'gp': 'Custom Gallery Page',
        'g': 'Gallery View'
    }



# coding=utf-8
from kntgen.file_helpers import *
from kntgen.json_to_dart.json_cmd import JsonCommand
from kntgen.model.network import *
from kntgen.printer import beauty_printer
from kntgen.str_helpers import *
from kntgen.template import Template


@auto_str
class NetworkTemplate(Template):
    def __init__(self, package_name: str, service: RestfulService):
        super().__init__('network')
        self.name_lower = create_file_name(service.name)
        self.package_name = package_name
        self.service = service
        self.generating_api = None
        self.append_file = False
        self.content_print_out = []

    def extract_content_from_template(self, template):
        return template.render(
            service=self.service,
            package_name=self.package_name,
            api=self.generating_api
        )

    def create_files(self, append_file: bool):
        self.append_file = append_file or False
        self._create_api_service()
        self._create_cloud_repo()
        self._create_data_cloud_repo()
        self._create_model_files()
        self._create_use_case_files()
        self._create_fixture_files()
        beauty_printer(title='Successfully created network related files',
                       content=self.content_print_out)

    def _create_api_service(self):
        scene_name = 'network'
        file_path, additional_info = self.create_file_from_template(
            class_name='api_service',
            template_file=f'{scene_name}.dart',
            sub_folder_template='service',
            output_path='lib/data/network',
            append_to_file=self.append_file,
            fixed_header=IMPORT_API_SERVICE_HEADER
        )
        self.content_print_out.append(
            make_content_output_file(file_path, additional_info))

    def _create_cloud_repo(self):
        scene_name = 'cloud_repo'
        file_path, additional_info = self.create_file_from_template(
            class_name=scene_name,
            template_file=f'{scene_name}.dart',
            sub_folder_template='repo',
            output_path='lib/repo',
            append_to_file=self.append_file,
            fixed_header=IMPORT_CLOUD_REPO_HEADER,
            fixed_footer=IMPORT_CLOUD_REPO_FOOTER
        )
        self.content_print_out.append(
            make_content_output_file(file_path, additional_info))

    def _create_data_cloud_repo(self):
        scene_name = 'data_cloud_repo'
        file_path, additional_info = self.create_file_from_template(
            class_name=scene_name,
            template_file=f'{scene_name}.dart',
            sub_folder_template='repo',
            output_path='lib/data/network/repo',
            append_to_file=self.append_file,
            fixed_header=IMPORT_DATA_CLOUD_REPO_HEADER,
            fixed_footer=IMPORT_DATA_CLOUD_REPO_FOOTER
        )
        self.content_print_out.append(
            make_content_output_file(file_path, additional_info))

    def _create_model_files(self):
        for index, api in enumerate(self.service.apis):
            self.generating_api = api
            # Request models
            file_path, additional_info = self.create_file_from_template(
                class_name=api.request_type_snake_case,
                template_file='request_model.dart',
                sub_folder_template='model',
                output_path='lib/model/network/request'
            )
            self.content_print_out.append(
                make_content_output_file(file_path, additional_info))
            # Add request file to [network_models.dart] exportation
            create_file(
                content=f'export \'network/request/{api.request_type_snake_case}.dart\';\n',
                file_name='network_models',
                file_extension='dart',
                folder='./lib/model',
                append_to_file=True
            )
            # Response models
            file_path, additional_info = JsonCommand(
                model_name=api.response_type_pascal_case,
                options={}
            ).create_files(
                json_str=api.response_json,
                folder='./lib/model/network/response',
                need_beauty_print=False
            )
            self.content_print_out.append(
                make_content_output_file(file_path, additional_info))
            # Add response file to [network_models.dart] exportation
            file_path, additional_info = create_file(
                content=f'export \'network/response/{api.response_type_snake_case}.dart\';\n',
                file_name='network_models',
                file_extension='dart',
                folder='./lib/model',
                append_to_file=True
            )
            if index == len(self.service.apis) - 1:
                self.content_print_out.append(
                    make_content_output_file(file_path, additional_info))

    def _create_use_case_files(self):
        # Use cases
        file_path, additional_info = self.create_file_from_template(
            class_name='base_use_case',
            template_file='base_use_case.dart',
            sub_folder_template='use_case',
            output_path='lib/use_case/core',
            append_to_file=True
        )
        self.content_print_out.append(
            make_content_output_file(file_path, additional_info))
        for index, api in enumerate(self.service.apis):
            use_case_name = f'{api.name_snake_case}_use_case'
            self.generating_api = api
            file_path, additional_info = self.create_file_from_template(
                class_name=use_case_name,
                template_file='network_use_case.dart',
                sub_folder_template='use_case',
                output_path='lib/use_case'
            )
            self.content_print_out.append(
                make_content_output_file(file_path, additional_info))
            # Add to [use_cases.dart] exportation
            file_path, additional_info = create_file(
                content=f'export \'{use_case_name}.dart\';\n',
                file_name='use_cases',
                file_extension='dart',
                folder='./lib/use_case',
                append_to_file=True
            )
            if index == len(self.service.apis) - 1:
                self.content_print_out.append(
                    make_content_output_file(file_path, additional_info))
            # Test Use cases
            file_path, additional_info = self.create_file_from_template(
                class_name=f'{use_case_name}_test',
                template_file='network_use_case_test.dart',
                sub_folder_template='use_case',
                output_path='test/use_case'
            )
            self.content_print_out.append(
                make_content_output_file(file_path, additional_info))

    def _create_fixture_files(self):
        file_path, additional_info = self.create_file_from_template(
            class_name='fixture_reader',
            template_file='fixture_reader.dart',
            sub_folder_template='other',
            output_path='test/fixture'
        )
        self.content_print_out.append(
            make_content_output_file(file_path, additional_info))
        for api in self.service.apis:
            file_path, additional_info = create_file(
                content=api.response_json,
                file_name=f'{api.response_type_snake_case}',
                file_extension='json',
                folder='./test/fixture'
            )
            self.content_print_out.append(
                make_content_output_file(file_path, additional_info))



# coding=utf-8

class DefinedCmd:
    HELP_CMD_SAMPLE = 'Print sample input of template files'
    HELP_CMD_BLOC = 'Create template files for a BLoC'
    HELP_CMD_MODEL = 'Create a model file from a JSON format'
    HELP_CMD_NETWORK = 'Create a Restful API service'
    HELP_CMD_UI = 'Generate Flutter component from Figma design link'

    class Argument:
        HELP_EXTENSION = 'extension of sample file'
        HELP_NAME_MODEL = 'object name of the JSON file'
        HELP_APPEND_FILE = 'append to the existed file'
        HELP_UI_COMPONENT = 'type of UI component we want to generate'

        @staticmethod
        def help_source(target: str):
            return f'source for {target}, support [pasteboard] ' \
                   'or [file path], default is [pasteboard]'

        @staticmethod
        def help_path(target: str):
            return f'path to the {target} file'

    class Required:
        INPUT_FILE = '[file] argument is required in file input option'


SUPPORTED_SOURCES = [
    'pb',
    'file'
]

POSTMAN_SOURCE = 'postman'

SUPPORTED_TEMPLATES = [
    'bloc',
    'model',
    'network',
    'ui'
]

SUPPORTED_SOURCE_FORMAT = [
    'json',
    'yaml'
]

SUPPORTED_UI_COMPONENT = [
    'string',
    'color',
    'image',
    'font',
    'widget',
    'enum',
    'translate'
]

DATE_REGEX = r'(\d{4})[-/](\d{2})[-/](\d{2})'

PYTHON_TO_DART_TYPES = {
    'int': 'num',
    'float': 'num',
    'bool': 'bool',
    'str': 'String',
    'NoneType': 'dynamic',
    'None': 'dynamic'
}

DART_TYPES_DEFAULT_VALUES = {
    'int': '0',
    'double': '0.0',
    'num': '0',
    'bool': 'false',
    'String': '',
    'DateTime?': 'null',
    'dynamic': 'null',
}

DART_TYPES_MOCK_VALUES = {
    'Int': '1',
    'Bool': 'true',
    'Double': '1.0',
    'String': '"foobar"',
    'Date': 'DateTime()',
    'dynamic': 'null',
}



from kntgen.cmd import Command, CommandOption
from kntgen.constants import *
from kntgen.file_helpers import get_flutter_project_package
from kntgen.model.exception import *
from kntgen.model.network import RestfulService
from kntgen.model.postman import PostmanCollection
from kntgen.network.network_template import NetworkTemplate


class NetworkCommand(Command):
    def __init__(self, options):
        super(NetworkCommand, self).__init__()
        # For future usage
        self.options = CommandOption(options)
        self.project_package = get_flutter_project_package()
        self.append_file = False

    @staticmethod
    def convert_postman_to_generate_form(postman_info: dict) -> RestfulService:
        postman_collection = PostmanCollection(**postman_info)
        return postman_collection.convert_to_restful_api_model()

    def create_files(self,
                     content_file: str,
                     file_extension: str,
                     append_file: bool):
        self.append_file = append_file or False
        try:
            if file_extension == SUPPORTED_SOURCE_FORMAT[0]:
                self._create_files_from_json(content_file)
            elif file_extension == SUPPORTED_SOURCE_FORMAT[1]:
                self._create_files_from_yaml(content_file)
        except Exception as e:
            DefinedErrorMessage.print_error(e)
            exit(1)

    def _create_files_from_json(self, json_str: str):
        service = NetworkTemplate(
            self.project_package,
            RestfulService.from_json(json_str)
        )
        service.create_files(self.append_file)

    def _create_files_from_yaml(self, yaml_stream):
        service = NetworkTemplate(
            self.project_package,
            RestfulService.from_yaml(yaml_stream)
        )
        service.create_files(self.append_file)



from kntgen.ui.figma import *
from kntgen.ui.flutter_parser import *


class FigmaUIGenerator:
    REGEX_FIGMA_LINK = r'(https:\/\/www.figma.com\/file\/)(?P<file_id>[A-Za-z0-9-]+)(\/\S.*\/?node-id=(?P<node_id>\S.*))?'

    def __init__(self, config: UiConfig):
        self.ui_config = config
        self.figma_config = config.figma
        self.google_config = config.google
        self.assets_config = config.assets
        self.file_id = ''
        self.node_id = ''
        self.figma = None
        self.file_node = None
        self.parent_node = None
        self._input_figma_source()

    def export_strings(self, need_print=True):
        parser = FStringParser(self.parent_node, self.figma, self.ui_config)
        parser.generate(folder=self.assets_config.translations,
                        pivot_locale=self.google_config.translate.main_locale,
                        need_print=need_print)

    def export_colors(self, need_print=True):
        parser = FColorParser(self.parent_node, self.figma, self.ui_config)
        parser.generate(folder=self.assets_config.colors,
                        need_print=need_print)

    def export_images(self, need_print=True):
        parser = FImageParser(self.parent_node, self.figma, self.ui_config)
        parser.generate(folder=self.assets_config.images,
                        svg_maximum_size=self.figma_config.maximum_svg_size_in_px,
                        need_print=need_print)

    def export_fonts(self, need_print=True):
        parser = FFontParser(self.parent_node, self.figma, self.ui_config)
        parser.generate(folder=self.assets_config.fonts,
                        api_key=self.google_config.font.api_key,
                        need_print=need_print)

    def export_widgets(self, need_preview=True):
        parser = FWidgetParser(self.parent_node, self.figma, self.ui_config)
        parser.generate(need_preview=need_preview,
                        gen_string=True,
                        gen_color=True,
                        gen_image=True,
                        gen_font=True)

    def export_enum(self, need_print=True):
        pass

    @staticmethod
    def translate_locale(folder: str = None,
                         pivot_locale: str = None,
                         placeholder: str = None,
                         need_print=True):
        FGenerator.translate_locales(locale_folder_path=folder,
                                     pivot_locale=pivot_locale,
                                     placeholder=placeholder,
                                     need_print=need_print)

    def _input_figma_source(self):
        def _print_finding_node_error(message=None):
            error_printer(message if message is not None
                          else f'Could not fetch Node information with id {self.node_id}!')
            exit(1)

        self.figma = FigmaPy(self.figma_config.access_token)
        self.file_id = self.figma_config.file_id
        self.node_id = self.figma_config.node_id

        if self.node_id is None or self.node_id.isspace():
            _print_finding_node_error(
                message='🙀 Sorry! No file id found!'
            )

        if self.node_id is None or self.node_id.isspace():
            _print_finding_node_error(
                message='🙀 Sorry! No node id found!'
            )

        info_printer(
            f'Fetching Figma File with file [{self.file_id}] node [{self.node_id}] ...')
        self.file_node = self.figma.get_file_nodes(
            file_key=self.file_id,
            ids=[self.node_id]
        )

        # Validate node information
        if self.node_id not in self.file_node.nodes:
            _print_finding_node_error()
        node_info = self.file_node.nodes[self.node_id]
        if node_info is None or 'document' not in node_info:
            _print_finding_node_error()
        self.parent_node = node_info['document']
        if self.parent_node is None or 'children' not in self.parent_node:
            _print_finding_node_error(
                message=f'Node {self.node_id} has no children!'
            )
        info_printer('Fetched Figma File successfully!')



import json

import yaml
from pydantic import BaseModel

from kntgen.str_helpers import *


@auto_str
class Param(BaseModel):
    name: str = None
    type_name: str = None
    default_value: str = None


@auto_str
class Unit(BaseModel):
    name: str = None
    params: list[Param] = []


@auto_str
class Event(Unit):
    need_check_connection: bool = False
    emit: list[str] = []


@auto_str
class State(Unit):
    is_error: bool = False


@auto_str
class Bloc(BaseModel):
    name: str = None
    need_singleton: bool = False
    need_cloud: bool = False
    need_custom_error: bool = False
    events: list[Event] = []
    states: list[State] = []

    @classmethod
    def from_json(cls, json_str: str):
        json_model = json.loads(json_str)
        return cls(**json_model)

    @classmethod
    def from_yaml(cls, yaml_stream):
        yaml_model = yaml.safe_load(yaml_stream)
        return cls(**yaml_model)

    @property
    def bloc_name_snake_case(self):
        return to_snake_case(self.name)



# API Scope Resource Objects
import math
from enum import Enum

from kntgen.str_helpers import *


# -------------------------------------------------------------------------
# FIGMA FILE PROPERTIES
# -------------------------------------------------------------------------
class File:
    # JSON file contents from a file
    def __init__(self, name, document, components, last_modified, thumbnail_url,
                 schema_version, styles):
        self.name = name  # File name
        self.last_modified = last_modified  # Date file was last modified
        self.thumbnail_url = thumbnail_url  # File thumbnail URL
        self.document = document  # Document content from a file
        self.components = components  # Document components from a file
        self.schema_version = schema_version  # Schema version from a file
        self.styles = styles  # Styles contained within a file


class FileNode:
    # JSON file contents from a file node
    def __init__(self, name, version, last_modified, thumbnail_url,
                 role, nodes):
        self.name = name  # File name
        self.last_modified = last_modified  # Date file was last modified
        self.thumbnail_url = thumbnail_url  # File thumbnail URL
        self.version = version  # Version of this node
        self.role = role  # Role of user in this file
        self.nodes = nodes  # Children of this node


class FileImages:
    # URLs for server-side rendered images from a file
    def __init__(self, images, err):
        self.err = err  # Error type as enum string
        self.images = images  # URLs of server-side rendered images from a file


class FileVersions:
    # Version history from a file
    def __init__(self, versions, pagination):
        self.versions = versions  # Version from a file
        self.pagination = pagination  # Pagination from a file


class Comments:
    # Comment(s) from a file
    def __init__(self, comments):
        self.comments = None  # Comment(s) from a file

        if len(comments) > 0 or comments is not None:
            self.comments = []
            for comment in comments:
                self.comments.append(Comment(comment['id'], comment['file_key'],
                                             comment['parent_id'],
                                             comment['user'],
                                             comment['created_at'],
                                             comment['resolved_at'],
                                             comment['message'],
                                             comment['client_meta'],
                                             comment['order_id']))


class TeamProjects:
    # Projects from a team
    def __init__(self, projects):
        self.projects = projects  # Projects from a team

    def get_project_name_by_id(self, id):
        for project in self.projects:
            if project['id'] == id:
                return project['name']

    def get_project_id_by_name(self, name):
        for project in self.projects:
            if project['name'] == name:
                return project['id']


class ProjectFiles:
    # Files from a project
    def __init__(self, files):
        self.files = files  # Files from a project


# -------------------------------------------------------------------------
# NODE PROPERTIES
# -------------------------------------------------------------------------
class Document:
    # The root node
    def __init__(self, children):
        self.children = children  # An array of canvases attached to the document


class Canvas:
    # Represents a single page
    def __init__(self, children, background_color, export_settings=None):
        self.children = children  # An array of top level layers on the canvas
        self.background_color = background_color  # Background color of the canvas
        self.export_settings = export_settings  # An array of export settings representing images to export. Default: []


class Frame:
    # A node of fixed size containing other nodes
    def __init__(self, children, background_color, blend_mode, constraints,
                 abs_bounding_box, size,
                 relative_transform, clips_content, preserve_ratio=False,
                 transition_node_id=None, opacity=1,
                 layout_grids=None, effects=None, is_mask=False,
                 export_settings=None):
        self.children = children  # An array of nodes that are direct children of this node
        self.background_color = background_color  # Background color of the node
        self.blend_mode = blend_mode  # How this node blends with nodes behind it in the scene
        self.constraints = constraints  # Horizontal and vertical layout constraints for node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element. Only present if geometry=paths is passed
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry.
        self.clips_content = clips_content  # Does this node clip content outside of its bounds?
        self.preserve_ratio = preserve_ratio  # Keep height and width constrained to same ratio
        self.transition_node_id = transition_node_id  # Node ID of node to transition to in prototyping
        self.opacity = opacity  # Opacity of the node
        self.layout_grids = layout_grids  # An array of layout grids attached to this node
        self.effects = effects  # An array of effects attached to this node
        self.is_mask = is_mask  # Does this node mask sibling nodes in front of it?
        self.export_settings = export_settings  # An array of export settings representing images to export from node


class Group:
    # A logical grouping of nodes [Holds properties of Frame except for layout_grids]
    def __init__(self, children, background_color, blend_mode, constraints,
                 abs_bounding_box, size,
                 relative_transform, clips_content, preserve_ratio=False,
                 transition_node_id=None, opacity=1,
                 effects=None, is_mask=False, export_settings=None):
        self.children = children  # An array of nodes that are direct children of this node
        self.background_color = background_color  # Background color of the node
        self.blend_mode = blend_mode  # How this node blends with nodes behind it in the scene
        self.constraints = constraints  # Horizontal and vertical layout constraints for node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element. Only present if geometry=paths is passed
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry.
        self.clips_content = clips_content  # Does this node clip content outside of its bounds?
        self.preserve_ratio = preserve_ratio  # Keep height and width constrained to same ratio
        self.transition_node_id = transition_node_id  # Node ID of node to transition to in prototyping
        self.opacity = opacity  # Opacity of the node
        self.effects = effects  # An array of effects attached to this node
        self.is_mask = is_mask  # Does this node mask sibling nodes in front of it?
        self.export_settings = export_settings  # An array of export settings representing images to export from node


class Vector:
    # A vector network, consisting of vertices and edges
    def __init__(self, blend_mode, constraints, abs_bounding_box, size,
                 relative_transform, fill_geometry,
                 stroke_weight, stroke_geometry, stroke_align,
                 export_settings=None, preserve_ratio=False,
                 transition_node_id=None, opacity=1, effects=None,
                 is_mask=False, fills=None, strokes=None):
        self.export_settings = export_settings  # An array of export settings representing images to export from node
        self.blend_mode = blend_mode  # How this node blends with nodes behind it in the scene
        self.preserve_ratio = preserve_ratio  # Keep height and width constrained to same ratio
        self.constraints = constraints  # Horizontal and vertical layout constraints for node
        self.transition_node_id = transition_node_id  # Node ID of node to transition to in prototyping
        self.opacity = opacity  # Opacity of the node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element. Only present if geometry=paths is passed
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry.
        self.effects = effects  # An array of effects attached to this node
        self.is_mask = is_mask  # Does this node mask sibling nodes in front of it?
        self.fills = fills  # An array of fill paints applied to the node
        self.fill_geometry = fill_geometry  # An array of paths representing the object fill
        self.strokes = strokes  # An array of stroke paints applied to the node
        self.stroke_weight = stroke_weight  # The weight of strokes on the node
        self.stroke_geometry = stroke_geometry  # An array of paths representing the object stroke
        self.stroke_align = stroke_align  # Where stroke is drawn relative to vector outline as a string enum


class Boolean:
    # A vector network, consisting of vertices and edges
    def __init__(self, children, blend_mode, constraints, abs_bounding_box,
                 size, relative_transform, fill_geometry,
                 stroke_weight, stroke_geometry, stroke_align,
                 export_settings=None, preserve_ratio=False,
                 transition_node_id=None, opacity=1, effects=None,
                 is_mask=False, fills=None, strokes=None):
        self.children = children  # An array of nodes that are being boolean operated on
        self.export_settings = export_settings  # An array of export settings representing images to export from node
        self.blend_mode = blend_mode  # How this node blends with nodes behind it in the scene
        self.preserve_ratio = preserve_ratio  # Keep height and width constrained to same ratio
        self.constraints = constraints  # Horizontal and vertical layout constraints for node
        self.transition_node_id = transition_node_id  # Node ID of node to transition to in prototyping
        self.opacity = opacity  # Opacity of the node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element. Only present if geometry=paths is passed
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry.
        self.effects = effects  # An array of effects attached to this node
        self.is_mask = is_mask  # Does this node mask sibling nodes in front of it?
        self.fills = fills  # An array of fill paints applied to the node
        self.fill_geometry = fill_geometry  # An array of paths representing the object fill
        self.strokes = strokes  # An array of stroke paints applied to the node
        self.stroke_weight = stroke_weight  # The weight of strokes on the node
        self.stroke_geometry = stroke_geometry  # An array of paths representing the object stroke
        self.stroke_align = stroke_align  # Where stroke is drawn relative to vector outline as a string enum


class Star:
    # A regular star shape [Shares properties of Vector]
    def __init__(self, blend_mode, constraints, abs_bounding_box, size,
                 relative_transform, fill_geometry,
                 stroke_weight, stroke_geometry, stroke_align,
                 export_settings=None, preserve_ratio=False,
                 transition_node_id=None, opacity=1, effects=None,
                 is_mask=False, fills=None, strokes=None):
        self.export_settings = export_settings  # An array of export settings representing images to export from node
        self.blend_mode = blend_mode  # How this node blends with nodes behind it in the scene
        self.preserve_ratio = preserve_ratio  # Keep height and width constrained to same ratio
        self.constraints = constraints  # Horizontal and vertical layout constraints for node
        self.transition_node_id = transition_node_id  # Node ID of node to transition to in prototyping
        self.opacity = opacity  # Opacity of the node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element. Only present if geometry=paths is passed
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry.
        self.effects = effects  # An array of effects attached to this node
        self.is_mask = is_mask  # Does this node mask sibling nodes in front of it?
        self.fills = fills  # An array of fill paints applied to the node
        self.fill_geometry = fill_geometry  # An array of paths representing the object fill
        self.strokes = strokes  # An array of stroke paints applied to the node
        self.stroke_weight = stroke_weight  # The weight of strokes on the node
        self.stroke_geometry = stroke_geometry  # An array of paths representing the object stroke
        self.stroke_align = stroke_align  # Where stroke is drawn relative to vector outline as a string enum


class Line:
    # A straight line [Shares properties of Vector]
    def __init__(self, blend_mode, constraints, abs_bounding_box, size,
                 relative_transform, fill_geometry,
                 stroke_weight, stroke_geometry, stroke_align,
                 export_settings=None, preserve_ratio=False,
                 transition_node_id=None, opacity=1, effects=None,
                 is_mask=False, fills=None, strokes=None):
        self.export_settings = export_settings  # An array of export settings representing images to export from node
        self.blend_mode = blend_mode  # How this node blends with nodes behind it in the scene
        self.preserve_ratio = preserve_ratio  # Keep height and width constrained to same ratio
        self.constraints = constraints  # Horizontal and vertical layout constraints for node
        self.transition_node_id = transition_node_id  # Node ID of node to transition to in prototyping
        self.opacity = opacity  # Opacity of the node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element. Only present if geometry=paths is passed
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry.
        self.effects = effects  # An array of effects attached to this node
        self.is_mask = is_mask  # Does this node mask sibling nodes in front of it?
        self.fills = fills  # An array of fill paints applied to the node
        self.fill_geometry = fill_geometry  # An array of paths representing the object fill
        self.strokes = strokes  # An array of stroke paints applied to the node
        self.stroke_weight = stroke_weight  # The weight of strokes on the node
        self.stroke_geometry = stroke_geometry  # An array of paths representing the object stroke
        self.stroke_align = stroke_align  # Where stroke is drawn relative to vector outline as a string enum


class Ellipse:
    # An ellipse [Shares properties of Vector]
    def __init__(self, blend_mode, constraints, abs_bounding_box, size,
                 relative_transform, fill_geometry,
                 stroke_weight, stroke_geometry, stroke_align,
                 export_settings=None, preserve_ratio=False,
                 transition_node_id=None, opacity=1, effects=None,
                 is_mask=False, fills=None, strokes=None):
        self.export_settings = export_settings  # An array of export settings representing images to export from node
        self.blend_mode = blend_mode  # How this node blends with nodes behind it in the scene
        self.preserve_ratio = preserve_ratio  # Keep height and width constrained to same ratio
        self.constraints = constraints  # Horizontal and vertical layout constraints for node
        self.transition_node_id = transition_node_id  # Node ID of node to transition to in prototyping
        self.opacity = opacity  # Opacity of the node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element. Only present if geometry=paths is passed
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry.
        self.effects = effects  # An array of effects attached to this node
        self.is_mask = is_mask  # Does this node mask sibling nodes in front of it?
        self.fills = fills  # An array of fill paints applied to the node
        self.fill_geometry = fill_geometry  # An array of paths representing the object fill
        self.strokes = strokes  # An array of stroke paints applied to the node
        self.stroke_weight = stroke_weight  # The weight of strokes on the node
        self.stroke_geometry = stroke_geometry  # An array of paths representing the object stroke
        self.stroke_align = stroke_align  # Where stroke is drawn relative to vector outline as a string enum


class RegularPolygon:
    # A regular n-sided polygon [Shares properties of Vector]
    def __init__(self, blend_mode, constraints, abs_bounding_box, size,
                 relative_transform, fill_geometry,
                 stroke_weight, stroke_geometry, stroke_align,
                 export_settings=None, preserve_ratio=False,
                 transition_node_id=None, opacity=1, effects=None,
                 is_mask=False, fills=None, strokes=None):
        self.export_settings = export_settings  # An array of export settings representing images to export from node
        self.blend_mode = blend_mode  # How this node blends with nodes behind it in the scene
        self.preserve_ratio = preserve_ratio  # Keep height and width constrained to same ratio
        self.constraints = constraints  # Horizontal and vertical layout constraints for node
        self.transition_node_id = transition_node_id  # Node ID of node to transition to in prototyping
        self.opacity = opacity  # Opacity of the node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element. Only present if geometry=paths is passed
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry.
        self.effects = effects  # An array of effects attached to this node
        self.is_mask = is_mask  # Does this node mask sibling nodes in front of it?
        self.fills = fills  # An array of fill paints applied to the node
        self.fill_geometry = fill_geometry  # An array of paths representing the object fill
        self.strokes = strokes  # An array of stroke paints applied to the node
        self.stroke_weight = stroke_weight  # The weight of strokes on the node
        self.stroke_geometry = stroke_geometry  # An array of paths representing the object stroke
        self.stroke_align = stroke_align  # Where stroke is drawn relative to vector outline as a string enum


class Rectangle:
    # A rectangle [Shares properties of Vector plus corner_radius]
    def __init__(self, corner_radius, blend_mode, constraints, abs_bounding_box,
                 size, relative_transform, fill_geometry,
                 stroke_weight, stroke_geometry, stroke_align,
                 export_settings=None, preserve_ratio=False,
                 transition_node_id=None, opacity=1, effects=None,
                 is_mask=False, fills=None, strokes=None):
        self.corner_radius = corner_radius  # Radius of each corner of the rectangle
        self.export_settings = export_settings  # An array of export settings representing images to export from node
        self.blend_mode = blend_mode  # How this node blends with nodes behind it in the scene
        self.preserve_ratio = preserve_ratio  # Keep height and width constrained to same ratio
        self.constraints = constraints  # Horizontal and vertical layout constraints for node
        self.transition_node_id = transition_node_id  # Node ID of node to transition to in prototyping
        self.opacity = opacity  # Opacity of the node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element. Only present if geometry=paths is passed
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry.
        self.effects = effects  # An array of effects attached to this node
        self.is_mask = is_mask  # Does this node mask sibling nodes in front of it?
        self.fills = fills  # An array of fill paints applied to the node
        self.fill_geometry = fill_geometry  # An array of paths representing the object fill
        self.strokes = strokes  # An array of stroke paints applied to the node
        self.stroke_weight = stroke_weight  # The weight of strokes on the node
        self.stroke_geometry = stroke_geometry  # An array of paths representing the object stroke
        self.stroke_align = stroke_align  # Where stroke is drawn relative to vector outline as a string enum


class Text:
    # A regular n-sided polygon [Shares properties of Vector]
    # plus characters, style, character_style_overrides, and style_override_table
    def __init__(self, characters, style, character_style_overrides,
                 style_override_table,
                 blend_mode, constraints, abs_bounding_box, size,
                 relative_transform, fill_geometry,
                 stroke_weight, stroke_geometry, stroke_align,
                 export_settings=None, preserve_ratio=False,
                 transition_node_id=None, opacity=1, effects=None,
                 is_mask=False, fills=None, strokes=None):
        self.characters = characters  # Text contained within text box
        self.style = style  # Style of text including font family and weight
        self.character_style_overrides = character_style_overrides  # Array with same number of elements as characters
        self.style_override_table = style_override_table  # Map from ID to TypeStyle for looking up style overrides
        self.export_settings = export_settings  # An array of export settings representing images to export from node
        self.blend_mode = blend_mode  # How this node blends with nodes behind it in the scene
        self.preserve_ratio = preserve_ratio  # Keep height and width constrained to same ratio
        self.constraints = constraints  # Horizontal and vertical layout constraints for node
        self.transition_node_id = transition_node_id  # Node ID of node to transition to in prototyping
        self.opacity = opacity  # Opacity of the node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element. Only present if geometry=paths is passed
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry.
        self.effects = effects  # An array of effects attached to this node
        self.is_mask = is_mask  # Does this node mask sibling nodes in front of it?
        self.fills = fills  # An array of fill paints applied to the node
        self.fill_geometry = fill_geometry  # An array of paths representing the object fill
        self.strokes = strokes  # An array of stroke paints applied to the node
        self.stroke_weight = stroke_weight  # The weight of strokes on the node
        self.stroke_geometry = stroke_geometry  # An array of paths representing the object stroke
        self.stroke_align = stroke_align  # Where stroke is drawn relative to vector outline as a string enum


class Slice:
    # A rectangular region of the canvas that can be exported
    def __init__(self, export_settings, abs_bounding_box, size,
                 relative_transform):
        self.export_settings = export_settings  # An array of export settings of images to export from this node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry


class Component:
    # A node that can have instances created of it that share the same properties
    def __init__(self, children, background_color, blend_mode, constraints,
                 abs_bounding_box, size,
                 relative_transform, clips_content, preserve_ratio=False,
                 transition_node_id=None, opacity=1,
                 layout_grids=None, effects=None, is_mask=False,
                 export_settings=None):
        self.children = children  # An array of nodes that are direct children of this node
        self.background_color = background_color  # Background color of the node
        self.blend_mode = blend_mode  # How this node blends with nodes behind it in the scene
        self.constraints = constraints  # Horizontal and vertical layout constraints for node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element. Only present if geometry=paths is passed
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry.
        self.clips_content = clips_content  # Does this node clip content outside of its bounds?
        self.preserve_ratio = preserve_ratio  # Keep height and width constrained to same ratio
        self.transition_node_id = transition_node_id  # Node ID of node to transition to in prototyping
        self.opacity = opacity  # Opacity of the node
        self.layout_grids = layout_grids  # An array of layout grids attached to this node
        self.effects = effects  # An array of effects attached to this node
        self.is_mask = is_mask  # Does this node mask sibling nodes in front of it?
        self.export_settings = export_settings  # An array of export settings representing images to export from node


class Instance:
    # An instance of a component, changes to the component result in the same changes applied to the instance
    def __init__(self, children, background_color, blend_mode, constraints,
                 abs_bounding_box, size, component_id,
                 relative_transform, clips_content, preserve_ratio=False,
                 transition_node_id=None, opacity=1,
                 layout_grids=None, effects=None, is_mask=False,
                 export_settings=None):
        self.component_id = component_id  # ID of component that this instance came from - refers to components table
        self.children = children  # An array of nodes that are direct children of this node
        self.background_color = background_color  # Background color of the node
        self.blend_mode = blend_mode  # How this node blends with nodes behind it in the scene
        self.constraints = constraints  # Horizontal and vertical layout constraints for node
        self.absolute_bounding_box = abs_bounding_box  # Bounding box of the node in absolute space coordinates
        self.size = size  # Width and height of element. Only present if geometry=paths is passed
        self.relative_transform = relative_transform  # Use to transform coordinates in geometry.
        self.clips_content = clips_content  # Does this node clip content outside of its bounds?
        self.preserve_ratio = preserve_ratio  # Keep height and width constrained to same ratio
        self.transition_node_id = transition_node_id  # Node ID of node to transition to in prototyping
        self.opacity = opacity  # Opacity of the node
        self.layout_grids = layout_grids  # An array of layout grids attached to this node
        self.effects = effects  # An array of effects attached to this node
        self.is_mask = is_mask  # Does this node mask sibling nodes in front of it?
        self.export_settings = export_settings  # An array of export settings representing images to export from node


# -------------------------------------------------------------------------
# FILE FORMAT TYPES
# -------------------------------------------------------------------------
class Color:
    # An RGBA Color
    def __init__(self, r, g, b, a):
        self.r = math.floor(r * 255)
        self.g = math.floor(g * 255)
        self.b = math.floor(b * 255)
        self.a = math.floor(a * 255)

    @staticmethod
    def valid_color_properties(color_properties):
        if 'r' not in color_properties \
                or 'g' not in color_properties \
                or 'b' not in color_properties:
            return False
        return True

    @staticmethod
    def rbg_to_hex(*t_color):
        return '#%02x%02x%02x' % t_color

    @staticmethod
    def hex_to_rgb(h_color: str):
        just_number_hex = h_color.lstrip('#')
        return tuple(int(just_number_hex[index:index + 2], 16)
                     for index in (0, 2, 4))


class ExportSetting:
    # Format and size to export an asset at
    def __init__(self, suffix, format, constraint):
        self.suffix = suffix  # File suffix to append all file names
        self.format = format  # Image type, string enum that supports values 'JPG', 'PNG', and 'SVG'
        self.constraint = constraint  # Constraint that determines sizing of exported asset


class Constraint:
    # Sizing constraint for exports
    def __init__(self, type, value):
        self.type = type  # Type of constraint to apply; string enum with potential values 'SCALE', 'WIDTH', 'HEIGHT'
        self.value = value  # See type property for effect of this field


class Rect:
    # A rectangle that expresses a bounding box in absolute coordinates
    def __init__(self, x, y, width, height):
        self.x = x  # X coordinate of top left corner of the rectangle
        self.y = y  # Y coordinate of top left corner of the rectangle
        self.width = width  # Width of the rectangle
        self.height = height  # Height of the rectangle


class RectLTRB:
    def __init__(self, x, y, width, height):
        """
        A rectangle that expresses a bounding box in absolute coordinates by Left>Top>Right>Bottom

        O(x, y) → X

        ↓

        Y

        :param x: X coordinate of top left corner of the rectangle
        :param y: Y coordinate of top left corner of the rectangle
        :param width: Width of the rectangle
        :param height: Height of the rectangle
        """
        self.left = x
        self.top = y
        self.right = x + width
        self.bottom = y + height

    def __str__(self) -> str:
        return f'{self.left}, {self.top}, {self.right}, {self.bottom}'

    @classmethod
    def cloned(cls, other_rect):
        if not other_rect:
            return cls(0, 0, 0, 0)
        return cls(
            other_rect.left,
            other_rect.top,
            other_rect.width,
            other_rect.height
        )

    @property
    def width(self):
        return self.right - self.left

    @property
    def height(self):
        return self.bottom - self.top

    @property
    def center_x(self):
        return self.left + self.width / 2

    @property
    def center_y(self):
        return self.top + self.height / 2

    def is_background_of(self, outer_frame, bias_figma_pixel=2):
        if outer_frame.left < self.left - bias_figma_pixel \
                or outer_frame.left > self.left + bias_figma_pixel:
            return False
        if outer_frame.top < self.top - bias_figma_pixel \
                or outer_frame.top > self.top + bias_figma_pixel:
            return False
        if outer_frame.right < self.right - bias_figma_pixel \
                or outer_frame.right > self.right + bias_figma_pixel:
            return False
        if outer_frame.bottom < self.bottom - bias_figma_pixel \
                or outer_frame.bottom > self.bottom + bias_figma_pixel:
            return False
        return True

    def is_intersected_with(self, other_rect, bias_figma_pixel=10):
        """
        Check if this rect is intersect with other rect

        :param other_rect:
        :param bias_figma_pixel:
        :return:
        """
        if self.right < other_rect.left + bias_figma_pixel \
                or self.bottom < other_rect.top + bias_figma_pixel \
                or self.left > other_rect.right - bias_figma_pixel \
                or self.top > other_rect.bottom - bias_figma_pixel:
            return False
        return True

    def is_contained_by(self, other_rect, bias_figma_pixel=0):
        """
        Check if this rect is contained by other rect

        :param other_rect:
        :param bias_figma_pixel:
        :return:
        """
        if self.left > other_rect.left - bias_figma_pixel \
                and self.top > other_rect.top - bias_figma_pixel \
                and self.right < other_rect.right + bias_figma_pixel \
                and self.bottom < other_rect.bottom + bias_figma_pixel:
            return True
        return False


class FillType(Enum):
    # Enum describing type of blend with layers below
    SOLID = 'SOLID'
    GRADIENT_LINEAR = 'GRADIENT_LINEAR'
    GRADIENT_RADIAL = 'GRADIENT_RADIAL'
    GRADIENT_ANGULAR = 'GRADIENT_ANGULAR'
    GRADIENT_DIAMOND = 'GRADIENT_DIAMOND'
    IMAGE = 'IMAGE'
    EMOJI = 'EMOJI'


class BlendMode(Enum):
    # Enum describing how layer blends with layers below
    # This type is a string enum with the following possible values:

    # Normal Blends
    PASS_THROUGH = 'PASS_THROUGH'
    NORMAL = 'NORMAL'
    # Darken
    DARKEN = 'DARKEN'
    MULTIPLY = 'MULTIPLY'
    LINEAR_BURN = 'LINEAR_BURN'
    COLOR_BURN = 'COLOR_BURN'
    # Lighten
    LIGHTEN = 'LIGHTEN'
    SCREEN = 'SCREEN'
    LINEAR_DODGE = 'LINEAR_DODGE'
    COLOR_DODGE = 'COLOR_DODGE'
    # Contrast
    OVERLAY = 'OVERLAY'
    SOFT_LIGHT = 'SOFT_LIGHT'
    HARD_LIGHT = 'HARD_LIGHT'
    # Inversion
    DIFFERENCE = 'DIFFERENCE'
    EXCLUSION = 'EXCLUSION'
    # Component
    HUE = 'HUE'
    SATURATION = 'SATURATION'
    COLOR = 'COLOR'
    LUMINOSITY = 'LUMINOSITY'


class LayoutConstraint:
    # Layout constraint relative to containing Frame
    def __init__(self, vertical, horizontal):
        self.vertical = vertical  # Vertical constraint as an enum
        self.horizontal = horizontal  # Horizontal constraint as an enum


class LayoutGrid:
    # Guides to align and place objects within a frame
    def __init__(self, pattern, section_size, visible, color, alignment,
                 gutter_size, offset, count):
        self.pattern = pattern  # Orientatoin of the grid as a string enum
        self.section_size = section_size  # Width of column grid or height of row grid or square grid spacing
        self.visible = visible  # Is the grid currently visible?
        self.color = color  # Color of the grid
        # The following properties are only meaningful for directional grids (COLUMNS or ROWS)
        self.alignment = alignment  # Positioning of grid as a string enum
        self.gutter_size = gutter_size  # Spacing in between columns and rows
        self.offset = offset  # Spacing before the first column or row
        self.count = count  # Number of columns or rows


class Effect:
    # A visual effect such as a shadow or blur
    def __init__(self, type, visible, radius, color, blend_mode, offset):
        self.type = type  # Type of effect as a string enum
        self.visible = visible  # is the effect active?
        self.radius = radius  # Radius of the blur effect (applies to shadows as well)
        # The following properties are for shadows only:
        self.color = color  # The color of the shadow
        self.blend_mode = blend_mode  # Blend mode of the shadow
        self.offset = offset  # How far the shadow is projected in the x and y directions


class Paint:
    # A solid color, gradient, or image texture that can be applied as fills or strokes
    def __init__(self, type, color, gradient_handle_positions, gradient_stops,
                 scale_mode,
                 visible=True, opacity=1):
        self.type = type  # Type of paint as a string enum
        self.visible = visible  # Is the paint enabled?
        self.opacity = opacity  # Overall opacity of paint (colors within the paint can also have opacity values)
        self.color = color  # Solid color of the paint
        # For gradient paints:
        self.gradient_handle_positions = gradient_handle_positions  # Three vectors, each are pos in normalized space
        self.gradient_stops = gradient_stops  # Positions of key points along the gradient axis with the anchored colors
        # For image paints:
        self.scale_mode = scale_mode  # Image scaling mode


class Vector2d:
    # A 2d vector
    def __init__(self, x, y):
        self.x = x  # X coordinate of the vector
        self.y = y  # Y coordinate of the vector


class Transform:
    # A 2x3 2D affine transformation matrix
    def __init__(self, matrix):
        self.matrix = matrix  # Transformation matrix


class Path:
    # A vector path
    def __init__(self, path, winding_rule):
        self.path = path  # A sequence of path commands in SVG notation
        self.winding_rule = winding_rule  # Winding rule for the path, either 'EVENODD' or 'NONZERO'


class FrameOffset:
    # A relative offset within a frame
    def __init__(self, node_id, node_offset):
        self.node_id = node_id  # Unique id specifying the frame
        self.node_offset = node_offset  # 2d vector offset within the frame


class ColorStop:
    # A position color pair representing a gradient stop
    def __init__(self, position, color):
        self.position = position  # Value between 0 and 1 representing position along gradient axis
        self.color = color  # Color attached to corresponding position


@auto_str
class TextStyle:
    def __init__(self, **style_node):
        self.font_family = style_node[
            'fontFamily'] if 'fontFamily' in style_node else ''
        self.font_post_script_name = style_node[
            'fontPostScriptName'] if 'fontPostScriptName' in style_node else ''
        self.font_weight = math.floor(style_node[
                                          'fontWeight']) if 'fontWeight' in style_node else 400
        self.font_size = math.floor(style_node[
                                        'fontSize']) if 'fontSize' in style_node else 14
        self.line_height_px = math.floor(style_node[
                                             'lineHeightPx']) if 'lineHeightPx' in style_node else 18
        self.italic = style_node['italic'] if 'italic' in style_node else False
        self.fills = style_node['fills'] if 'fills' in style_node else {}

    @property
    def is_regular(self):
        return self.font_weight == 400

    @property
    def font_weight_in_google_format(self):
        return 'regular' if self.is_regular else str(self.font_weight)

    @property
    def italic_in_google_format(self):
        return 'italic' if self.italic else ''

    @staticmethod
    def extract_google_font_family(styles: list, basic_font=False):
        font_dict = {}
        for style in styles or []:
            f_family = style.font_family
            f_post_script = style.font_post_script_name
            if f_family not in font_dict:
                font_dict[f_family] = set()
            if basic_font:
                if not font_dict[f_family]:
                    font_dict[f_family].update(
                        [('regular', f'{f_family}-Regular')])
                continue
            if not any(f_post_script in f_font_tuple for f_font_tuple in
                       font_dict[f_family]):
                f_weight = style.font_weight_in_google_format
                f_italic = style.italic_in_google_format
                if style.is_regular:
                    if style.italic:
                        font_dict[f_family].add((f_italic, f_post_script))
                    else:
                        font_dict[f_family].add((f_weight, f_post_script))
                else:
                    font_dict[f_family].add(
                        (f'{f_weight}{f_italic}', f_post_script))
        return {family: [*style, ] for family, style in font_dict.items()}


class ComponentDescription:
    # A description of a master component. Helps you identify which component instances are attached to
    def __init__(self, name, description):
        self.name = name  # The name of the component
        self.description = description  # The description of the component as entered in the editor


# -------------------------------------------------------------------------
# GENERAL API TYPES
# -------------------------------------------------------------------------
class Comment:
    # A comment or reply left by a user
    def __init__(self, id, file_key, parent_id, user, created_at, resolved_at,
                 message, client_meta, order_id):
        self.id = id  # Unique identifier for comment
        self.file_key = file_key  # The file in which the comment lives
        self.parent_id = parent_id  # If present, the id of the comment to which this is the reply
        self.user = user  # The user who left the comment
        self.created_at = created_at  # The UTC ISO 8601 time at which the comment was left
        self.resolved_at = resolved_at  # If set, the UTC ISO 8601 time the comment was resolved
        self.message = message  # Content of comment
        self.client_meta = client_meta  # The position of the comment. Absolute coordinates or relative offset
        self.order_id = order_id  # Only set for top level comments. The number displayed with the comment in the UI


class User:
    # A description of a user
    def __init__(self, handle, img_url):
        self.handle = handle  # Name of the user
        self.img_url = img_url  # URL link to the user's profile image


class Version:
    # A version of a file
    def __init__(self, id, created_at, label, description, user):
        self.id = id  # Unique identifier for version
        self.created_at = created_at  # the UTC ISO 8601 time at which the version was created
        self.label = label  # The label given to the version in the editor
        self.description = description  # The description of the version as entered in the editor
        self.user = user  # The user that created the version



import json
from collections import OrderedDict
from typing import Any

import regex
import yaml
from pydantic import BaseModel

from kntgen.constants import *
from kntgen.str_helpers import *

DEFAULT_CONTENT_TYPE = 'String'
DEFAULT_CONTENT_TYPE_FILE = 'File'
REGEX_FIND_PATH = r'/{(.*?)}'
ENDPOINT_CONVERSION = {
    'get': '@GET',
    'post': '@POST',
    'patch': '@PATCH',
    'put': '@PUT',
    'delete': '@DELETE'
}
IMPORT_API_SERVICE_HEADER = """import 'package:dio/dio.dart' hide Headers;
import 'package:retrofit/http.dart';

import '../../model/network_models.dart';

part 'api_service.g.dart';
"""
IMPORT_CLOUD_REPO_HEADER = """import '../model/network_models.dart';

abstract class CloudRepo {
"""
IMPORT_CLOUD_REPO_FOOTER = """}
"""
IMPORT_DATA_CLOUD_REPO_HEADER = """import 'package:dio/dio.dart';
import 'package:injectable/injectable.dart';

import '../../../model/network_models.dart';
import '../../../repo/cloud_repo.dart';

@Singleton(as: CloudRepo)
class DataCloudRepo {"""
IMPORT_DATA_CLOUD_REPO_FOOTER = """
}

Future<Dio> _createDio({
  bool needAuthorized = false,
  Map<String, dynamic> customHeaders = const {},
}) async {
  final dio = Dio(BaseOptions(
    connectTimeout: _defaultConnectionTimeout,
    receiveTimeout: _defaultReceiveTimeout,
    headers: customHeaders,
  ));
  if (needAuthorized) {
    final token = '';
    dio.interceptors
        .add(InterceptorsWrapper(onRequest: (options, handlers) async {
      dio.lock();
      options.headers[HttpHeaders.authorizationHeader] = 'Bearer $token';
      options.headers[HttpHeaders.contentTypeHeader] =
      'application/json; charset=UTF-8';
      dio.unlock();
    }));
  }
  if (Flavors.debugMode) {
    dio.interceptors.add(PrettyDioLogger(
      request: true,
      requestHeader: true,
    ));
  }
  return dio;
}
"""


@auto_str
class Unit(BaseModel):
    title: str
    content: str = None
    content_type: str = DEFAULT_CONTENT_TYPE

    def __init__(self, **data: Any):
        super().__init__(**data)
        if self.content and is_number(self.content):
            self.content_type = 'num'

    @property
    def title_camel_case(self):
        just_letter = ''.join([c for c in self.title if c not in '-._'])
        return to_camel_case(just_letter)

    @property
    def is_file_type(self):
        return 'file' in self.content_type.lower()

    @property
    def is_list_type(self):
        return 'list' in self.content_type.lower()

    @property
    def is_optional(self):
        return self.content_type.endswith('?')

    @property
    def value(self):
        value = self.content
        if value and self.content_type.startswith('String'):
            value = f'{value!r}'
        return value

    @property
    def non_null_value(self):
        if self.content:
            value = self.content
        elif self.content_type in DART_TYPES_DEFAULT_VALUES:
            value = DART_TYPES_DEFAULT_VALUES[self.content_type]
        elif self.is_list_type:
            value = '[]'
        elif self.is_file_type:
            value = f'\'path_to_{self.title}\''
        else:
            value = f'{self.content_type}()'
        if self.content_type.startswith('String'):
            value = f'{value!r}'
        return value


@auto_str
class API(BaseModel):
    name: str
    method: str = 'get'
    endpoint: str
    request_type: str = None
    headers: list[Unit] = []
    queries: list[Unit] = []
    body: list[Unit] = []
    fields: list[Unit] = []
    parts: list[Unit] = []
    response_type: str = None
    response_json: str = '{}'

    def __init__(self, **data: Any):
        super().__init__(**data)
        # Remove mistake if has
        if self.method == 'get':
            self.body = []
            self.fields = []
            self.parts = []

    @property
    def has_param(self):
        return len(self.path_from_endpoint) > 0 \
               or len(self.headers) > 0 \
               or len(self.queries) > 0 \
               or len(self.body) > 0 \
               or len(self.fields) > 0 \
               or len(self.parts) > 0

    @property
    def name_camel_case(self):
        return to_camel_case(self.name)

    @property
    def name_pascal_case(self):
        return to_pascal_case(self.name)

    @property
    def name_snake_case(self):
        return to_snake_case(self.name)

    @property
    def name_to_word(self):
        return to_words(self.name)

    @property
    def request_type_pascal_case(self):
        return to_pascal_case(self.request_type)

    @property
    def request_type_snake_case(self):
        return to_snake_case(self.request_type)

    @property
    def response_type_pascal_case(self):
        return to_pascal_case(self.response_type)

    @property
    def response_type_snake_case(self):
        return to_snake_case(self.response_type)

    @property
    def is_get_method(self):
        return self.method == 'get'

    @property
    def retrofit_endpoint(self):
        method_type = ENDPOINT_CONVERSION[self.method] or '@GET'
        return f'{method_type}({self.endpoint!r})'

    @property
    def static_headers(self):
        return [header for header in self.headers if header.content]

    @property
    def dynamic_headers(self):
        return [header for header in self.headers if not header.content]

    @property
    def path_from_endpoint(self):
        # Extract paths if needed
        paths = regex.findall(REGEX_FIND_PATH, self.endpoint)
        return {path: to_camel_case(path) for path in paths}

    @property
    def has_fields(self):
        return len(self.fields) > 0

    @property
    def has_file_part(self):
        for part in self.parts:
            if part.is_file_type:
                return True
        return False


@auto_str
class RestfulService(BaseModel):
    name: str
    base_url: str
    headers: list[Unit] = []
    apis: list[API] = []

    @classmethod
    def from_json(cls, json_str: str):
        json_model = json.loads(json_str)
        return cls(**json_model)

    @classmethod
    def from_yaml(cls, yaml_stream):
        yaml_model = yaml.safe_load(yaml_stream)
        return cls(**yaml_model)

    @property
    def name_camel_case(self):
        return to_camel_case(self.name)

    @property
    def name_pascal_case(self):
        return to_pascal_case(self.name)

    @property
    def name_snake_case(self):
        return to_snake_case(self.name)

    def to_dict(self):
        return json.loads(json.dumps(self, default=lambda o: o.__dict__),
                          object_pairs_hook=OrderedDict)



# PYTHON_ARGCOMPLETE_OK
# -*- coding: UTF-8 -*-

from arghandler import subcmd, ArgumentHandler

from kntgen import __version__
from kntgen.bloc.bloc_cmd import BlocCommand
from kntgen.json_to_dart.json_cmd import JsonCommand
from kntgen.model.ui_config import UiConfig
from kntgen.network.network_cmd import NetworkCommand
from kntgen.parser_helpers import *
from kntgen.printer import *
from kntgen.ui.figma_ui_generator import FigmaUIGenerator


@subcmd('sample', help=DefinedCmd.HELP_CMD_SAMPLE)
def cmd_sample(parser, _, args):
    parser.description = DefinedCmd.HELP_CMD_SAMPLE
    parser.add_argument(
        'template',
        nargs=1,
        choices=SUPPORTED_TEMPLATES
    )
    parser.add_argument(
        '-e', '--extension',
        choices=SUPPORTED_SOURCE_FORMAT,
        nargs=1,
        required=False,
        help=DefinedCmd.Argument.HELP_EXTENSION
    )

    # Fetch arguments
    args = parser.parse_args(args)
    template = args.template[0]
    if template not in SUPPORTED_TEMPLATES:
        parser.error(DefinedErrorMessage.ERROR_SUPPORT_TEMPLATE)

    extension = SUPPORTED_SOURCE_FORMAT[0]
    if args.extension:
        extension = args.extension[0]
    if extension not in SUPPORTED_SOURCE_FORMAT:
        parser.error(DefinedErrorMessage.ERROR_SUPPORT_FORMAT)

    # Print the sample
    root = '../kntgen_templates'
    file_path = ''
    if template == SUPPORTED_TEMPLATES[0]:
        file_path = '/bloc/sample_input/sample_bloc'
    elif template == SUPPORTED_TEMPLATES[1]:
        file_path = '/json_to_dart/sample_input/sample'
    elif template == SUPPORTED_TEMPLATES[2]:
        file_path = '/network/sample_input/sample_network'
    elif template == SUPPORTED_TEMPLATES[3]:
        file_path = '/ui/sample_input/kntgen_config'
        # UI config only accept .yaml
        extension = SUPPORTED_SOURCE_FORMAT[1]
    file_path = root + file_path + '.' + extension
    try:
        # Open file depends on extension
        with open(file_path, 'r') as file:
            content_file = file.read()
        beauty_printer(title=f'Sample [{template}.{extension}] file',
                       content=[content_file],
                       content_has_block=False)
    except (IOError, OSError):
        print(DefinedErrorMessage.FILE_NOT_FOUND)
        exit(1)


@subcmd(SUPPORTED_TEMPLATES[0], help=DefinedCmd.HELP_CMD_BLOC)
def cmd_bloc(parser, _, args):
    create_base_argument(parser,
                         description=DefinedCmd.HELP_CMD_BLOC,
                         target='BLoC configuration')
    _, content_file, file_extension = handle_base_parser(parser, args)
    BlocCommand({}).create_files(content_file, file_extension)


@subcmd(SUPPORTED_TEMPLATES[1], help=DefinedCmd.HELP_CMD_MODEL)
def cmd_model(parser, _, args):
    parser.add_argument(
        'name',
        nargs=1,
        help=DefinedCmd.Argument.HELP_NAME_MODEL
    )
    create_base_argument(parser,
                         description=DefinedCmd.HELP_CMD_MODEL,
                         target='JSON')
    args, content_file, _ = handle_base_parser(parser, args)
    model_name = args.name[0]
    JsonCommand(model_name=model_name, options={}).create_files(content_file)


@subcmd(SUPPORTED_TEMPLATES[2], help=DefinedCmd.HELP_CMD_NETWORK)
def cmd_network(parser, _, args):
    parser.add_argument(
        '-a', '--append',
        required=False,
        action='store_true',
        help=DefinedCmd.Argument.HELP_APPEND_FILE
    )
    supported_sources = SUPPORTED_SOURCES.copy()
    supported_sources.append(POSTMAN_SOURCE)
    create_base_argument(parser,
                         description=DefinedCmd.HELP_CMD_NETWORK,
                         target='Restful API configuration',
                         supported_sources=supported_sources)
    extracted_args, content_file, file_extension = handle_base_parser(
        parser, args, supported_sources=supported_sources)
    has_append_option = is_set('-a') or is_set('--append')
    # Handle input Postman source
    source_type = extracted_args.source[0]
    if source_type == POSTMAN_SOURCE:
        file_extension = SUPPORTED_SOURCE_FORMAT[0]
        convert_folder = './kntgen'
        convert_file = 'network_config'
        convert_file_path = f'{convert_folder}/{convert_file}.{file_extension}'
        # Convert postman content into worked network form
        restful_service = NetworkCommand.convert_postman_to_generate_form(
            json.loads(content_file))
        generated_file_path, additional_info = create_json_file(
            restful_service.to_dict(),
            convert_file_path)
        generated_content = make_content_output_file(generated_file_path,
                                                     additional_info)
        beauty_printer(title='Successfully create network config file!',
                       content=[generated_content])
        # Save into a file and ask user to fill in
        info_printer(
            f'Please open file [{convert_file_path}] and fill required fields!')
        user_opt = input('Have you completed? (y/n)')
        if user_opt == 'y' or user_opt == 'Y':
            # Read network information from user edited file
            with open(convert_file_path, 'r', encoding='utf8') as edited_file:
                content_file = edited_file.read()
    # Start generating all related network files
    NetworkCommand({}).create_files(content_file,
                                    file_extension,
                                    has_append_option)


@subcmd(SUPPORTED_TEMPLATES[3], help=DefinedCmd.HELP_CMD_UI)
def cmd_ui(parser, _, args):
    parser.add_argument(
        '-c', '--component',
        nargs=1,
        choices=SUPPORTED_UI_COMPONENT,
        required=False,
        help=DefinedCmd.Argument.HELP_UI_COMPONENT
    )
    args = parser.parse_args(args)
    component = args.component[0] if args.component \
        else SUPPORTED_UI_COMPONENT[0]
    if component not in SUPPORTED_UI_COMPONENT:
        parser.error(DefinedErrorMessage.ERROR_SUPPORT_COMPONENT)

    # Firstly we check kntgen_config.yaml file to get configurations
    with open('./kntgen_config.yaml', 'r', encoding='utf8') as config_file:
        ui_config = UiConfig.from_yaml(config_file.read())

    # Then start parsing
    if component == SUPPORTED_UI_COMPONENT[6]:
        FigmaUIGenerator.translate_locale(
            folder=ui_config.assets.translations,
            pivot_locale=ui_config.google.translate.main_locale,
            placeholder=ui_config.google.translate.placeholder)
        return

    generator = FigmaUIGenerator(ui_config)
    if component == SUPPORTED_UI_COMPONENT[0]:
        generator.export_strings()
    elif component == SUPPORTED_UI_COMPONENT[1]:
        generator.export_colors()
    elif component == SUPPORTED_UI_COMPONENT[2]:
        generator.export_images()
    elif component == SUPPORTED_UI_COMPONENT[3]:
        generator.export_fonts()
    elif component == SUPPORTED_UI_COMPONENT[4]:
        generator.export_widgets()
    elif component == SUPPORTED_UI_COMPONENT[5]:
        generator.export_enum()


def main():
    handler = ArgumentHandler(
        use_subcommand_help=True,
        enable_autocompletion=True,
        epilog='Get help on a subcommand: kntgen subcommand -h'
    )

    handler.add_argument(
        '-v', '--version',
        action='version',
        version=__version__,
        help='show the version number'
    )

    # if no parameters are provided, show help
    if len(sys.argv) == 1:
        handler.run(['-h'])
    else:
        handler.run()


if __name__ == '__main__':
    main()



from enum import Enum

from kntgen.model.network import *
from kntgen.str_helpers import *

REQUIRED_FIELD = '**required**'


class HttpMethod(str, Enum):
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    PATCH = 'PATCH'

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_

    @property
    def value_lower_case(self):
        return self.value.lower()


class HttpBodyMode(str, Enum):
    RAW = 'raw'
    FORM_DATA = 'formdata'
    FORM_URL_ENCODED = 'urlencoded'

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


class HttpFormDataType(str, Enum):
    TEXT = 'text'
    FILE = 'file'

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_

    def convert_to_network_form_data_type(self):
        return {
            HttpFormDataType.TEXT: DEFAULT_CONTENT_TYPE,
            HttpFormDataType.FILE: DEFAULT_CONTENT_TYPE_FILE
        }.get(self, DEFAULT_CONTENT_TYPE)


@auto_str
class PostmanUnit:
    def __init__(self, **kwargs):
        self.key = kwargs.get('key')
        self.value = kwargs.get('value')
        self.type_value = kwargs.get('type')
        self.type = HttpFormDataType(self.type_value) \
            if HttpFormDataType.has_value(
            self.type_value) else HttpFormDataType.TEXT
        self.source_file = kwargs.get('src')

    def __eq__(self, other):
        return isinstance(other, __class__) \
               and other.key == self.key \
               and other.value == self.value

    def __hash__(self):
        return id(f'{self.key}_{self.value}')

    def convert_to_network_unit(self):
        return Unit(
            title=self.key,
            content=self.value,
            content_type=self.type.convert_to_network_form_data_type()
        )


@auto_str
class PostmanApiRequestBody:
    def __init__(self, **kwargs):
        self.mode_value = kwargs.get('mode')
        self.mode = HttpBodyMode(self.mode_value) \
            if HttpBodyMode.has_value(self.mode_value) else None
        self.data = kwargs.get(self.mode_value)

    def convert_to_body_network_unit(self):
        if self.mode == HttpBodyMode.RAW:
            data_dict = json.loads(self.data or '{}',
                                   object_pairs_hook=OrderedDict)
            units = []
            return Unit(
                title=self.key,
                content=self.value
            )
        return []

    def convert_to_field_network_unit(self):
        return Unit(
            title=self.key,
            content=self.value
        )

    def convert_to_part_network_unit(self):
        return Unit(
            title=self.key,
            content=self.value
        )


@auto_str
class PostmanApiRequest:
    def __init__(self, **kwargs):
        self.method = HttpMethod(kwargs.get('method')) \
            if HttpMethod.has_value(kwargs.get('method')) else HttpMethod.GET
        self.url = kwargs.get('url', {})
        self.headers = [PostmanUnit(**v) for v in kwargs.get('header', [])]
        self.body = PostmanApiRequestBody(
            **kwargs['body']) if 'body' in kwargs else None
        self.protocol = self.url.get('protocol')
        self.hosts = self.url.get('host', [])
        self.paths = [f'{{{p[1:]}}}' if p.startswith(':') else p for p in
                      self.url.get('path', []) if p]
        self.queries = [PostmanUnit(**v) for v in self.url.get('query', [])]

    @property
    def host(self):
        if not self.hosts:
            return REQUIRED_FIELD
        protocol_present = f'{self.protocol}://' if self.protocol else ''
        return protocol_present + '/'.join(self.hosts)

    @property
    def endpoint(self):
        if not self.paths:
            return REQUIRED_FIELD
        return '/' + '/'.join(self.paths)

    def body(self):
        return self.body.convert_to_body_network_unit() if self.body else []

    def fields(self):
        return self.body.convert_to_field_network_unit() if self.body else []

    def parts(self):
        return self.body.convert_to_part_network_unit() if self.body else []


@auto_str
class PostmanApi:
    def __init__(self, **kwargs):
        self.name = kwargs.get('name')
        self.sample_api = kwargs.get('response', [])
        self.request = None
        if self.sample_api:
            self.request = PostmanApiRequest(
                **self.sample_api[0]['originalRequest']) \
                if 'originalRequest' in self.sample_api[0] else None
        self.response = {}

    @property
    def name_camel_case(self):
        return to_camel_case(self.name)

    @property
    def name_pascal_case(self):
        return to_pascal_case(self.name)

    def method(self):
        return self.request.method.value_lower_case \
            if self.request else HttpMethod.GET.value_lower_case

    def host(self):
        return self.request.host if self.request else REQUIRED_FIELD

    def endpoint(self):
        return self.request.endpoint if self.request else REQUIRED_FIELD

    def request_type(self):
        return f'{self.name_pascal_case}Request'

    def headers(self):
        return [header_unit.convert_to_network_unit() for header_unit in
                self.request.headers] if self.request else []

    def queries(self):
        return [query_unit.convert_to_network_unit() for query_unit in
                self.request.queries] if self.request else []

    def response_type(self):
        return f'{self.name_pascal_case}Response'


@auto_str
class PostmanEnvironmentUnit:
    def __init__(self, **kwargs):
        self.key = kwargs.get('key')
        self.value = kwargs.get('value')
        self.enabled = kwargs.get('enabled', True)


@auto_str
class PostmanEnvironment:
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.name = kwargs.get('name')
        self.units = [PostmanEnvironmentUnit(**v) for v in
                      kwargs.get('values', [])]


@auto_str
class PostmanCollection:
    def __init__(self, **kwargs):
        info = kwargs.get('info', dict())
        self.id = info.get('_postman_id')
        self.name = info.get('name')
        self.schema = info.get('schema')
        self.environment = PostmanEnvironment(**info.get('environment')) \
            if 'environment' in info else None
        # Recursive to find all API from sub folder
        raw_apis = []
        PostmanCollection.collect_raw_api(postman_item=kwargs.get('item', []),
                                          results=raw_apis)
        self.apis = [PostmanApi(**i) for i in raw_apis]

    @staticmethod
    def collect_raw_api(postman_item, results: list):
        if postman_item is None or results is None:
            return
        if isinstance(postman_item, list):
            for sub_item in postman_item:
                PostmanCollection.collect_raw_api(
                    postman_item=sub_item,
                    results=results)
            return
        if not isinstance(postman_item, dict):
            return
        if 'response' in postman_item:
            results.append(postman_item)
            return
        if 'item' in postman_item:
            PostmanCollection.collect_raw_api(
                postman_item=postman_item['item'],
                results=results)

    @property
    def common_host(self):
        # Get common host
        host = REQUIRED_FIELD
        if self.apis:
            existed_host_apis = [api for api in self.apis if
                                 api.host != REQUIRED_FIELD]
            if existed_host_apis:
                first_api_host = existed_host_apis[0].host
                for api in existed_host_apis[1:]:
                    if first_api_host != api.host:
                        return host
                host = first_api_host
        # Then replace with environment value if needed
        if self.environment:
            for unit in self.environment.units:
                if unit.enabled and unit.key in host:
                    host = unit.value
                    if not ''.startswith('https://'):
                        host = f'https://{host}'
                    break
        return host

    @property
    def common_headers(self):
        # Find all common headers
        headers = []
        if self.apis:
            existed_header_apis = [api for api in self.apis if
                                   api.request and api.request.headers]
            if len(existed_header_apis) != len(self.apis):
                # Guaranteed this case doesn't have common headers
                return headers

            headers = existed_header_apis[0].request.headers
            for api in existed_header_apis[1:]:
                headers = list(set(api.request.headers).intersection(headers))
        return headers

    def convert_to_restful_api_model(self):
        return RestfulService(
            name=self.name,
            base_url=self.common_host,
            headers=[header_unit.convert_to_network_unit() for header_unit in
                     self.common_headers],
            apis=[
                API(
                    name=api.name_camel_case,
                    method=api.method(),
                    endpoint=api.endpoint(),
                    request_type=api.request_type(),
                    headers=api.headers(),
                    response_type=api.response_type()
                ) for api in self.apis if api
            ]
        )



from kntgen.str_helpers import *


@auto_str
class ProjectInfo:
    def __init__(self, project, developer, company, project_id):
        self.project = project
        self.developer = developer
        self.company = company
        self.project_id = project_id



import math
from multiprocessing import Process, Queue

from kntgen.printer import error_printer

MAX_PROCESS = 50


def run_parallel_tasks(task_executor, tasks: list, *args, **kwargs):
    if not tasks:
        error_printer('Not found any task!')
        return

    def _split_to_section_tasks():
        total_section_task = max(MAX_PROCESS, len(tasks))
        task_per_section = math.ceil(len(tasks) / total_section_task)
        results = []
        for i in range(total_section_task):
            start = i * task_per_section
            end = (i + 1) * task_per_section
            if end >= len(tasks):
                end = None
            results.append(tasks[start:end])
            if end is None:
                break
        return results

    result_tasks = []
    mp = Multiprocessor()
    section_tasks = _split_to_section_tasks()
    for section_task in section_tasks:
        mp.run(_run_section_task, task_executor, section_task, *args, **kwargs)
    # Wait for all processes to finish
    result_section_tasks = mp.wait()
    for result_section_task in result_section_tasks:
        result_tasks.extend(result_section_task)
    return result_tasks


def _run_section_task(task_executor, section, *args, **kwargs):
    results = []
    for task in section:
        results.append(task_executor(task, *args, **kwargs))
    return results


class Multiprocessor:
    def __init__(self):
        self.processes = []
        self.queue = Queue()

    def run(self, func, *args, **kwargs):
        args2 = [func, self.queue, args, kwargs]
        p = Process(target=self._wrapper, args=args2)
        self.processes.append(p)
        p.start()

    def wait(self):
        rets = []
        for _ in self.processes:
            ret = self.queue.get()
            rets.append(ret)
        for p in self.processes:
            p.join()
        return rets

    @staticmethod
    def _wrapper(func, queue, args, kwargs):
        ret = func(*args, **kwargs)
        queue.put(ret)



import json
import os
import sys
from collections import OrderedDict
from pathlib import Path

import requests
import yaml
from tqdm import tqdm


class ModifiedFileState:
    NO_CHANGE = '[Content already existed]'
    HAS_APPENDED = '[New content appended]'
    HAS_CREATED_NEW = '[New content generated]'


def download_to_file(url: str, file_path: str):
    response = requests.get(url, stream=True)
    total_size_in_bytes = int(response.headers.get('content-length', 0))
    block_size = 1024
    *_, file_name = os.path.split(file_path)
    progress_bar = tqdm(
        total=total_size_in_bytes,
        desc=f'Downloading {file_name}',
        bar_format='{desc:50} {bar:30} {percentage:3.0f}% [{n_fmt}/{total_fmt}, {rate_fmt}]',
        unit='iB',
        unit_scale=True,
        colour='blue'
    )
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    with open(file_path, 'wb') as file:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            file.write(data)
    progress_bar.close()
    if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
        print('ERROR, something went wrong')
        return None
    return file_path


def create_file(content,
                file_name,
                file_extension,
                folder=None,
                append_to_file=False,
                fixed_header: str = None,
                fixed_footer: str = None):
    if folder:
        if file_extension:
            file_path = '{}/{}.{}'.format(folder, file_name, file_extension)
        else:
            file_path = '{}/{}'.format(folder, file_name)
    else:
        if file_extension:
            file_path = '{}.{}'.format(file_name, file_extension)
        else:
            file_path = '{}'.format(file_name)
    # write to file
    try:
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        content_file = ''
        additional_info = ''
        if Path(file_path).exists():
            with open(file_path, 'r') as file:
                content_file = file.read()
            if append_to_file is True and content in content_file:
                additional_info = ModifiedFileState.NO_CHANGE
                return file_path, additional_info
        with open(file_path, 'wb') as file:
            header = fixed_header or ''
            footer = fixed_footer or ''
            if append_to_file is not True:
                content = header + content + footer
                additional_info = ModifiedFileState.HAS_CREATED_NEW
            else:
                if header not in content_file:
                    content = header + content
                content_file = content_file.replace(footer, '')
                content = content_file + content + footer
                additional_info = ModifiedFileState.HAS_APPENDED
            file.write(content.encode('utf8'))
    except Exception as e:
        print(e)
    return file_path, additional_info


def create_json_file(json_dict: dict,
                     file_path):
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    with open(file_path, 'w', encoding='utf8') as json_file:
        json.dump(json_dict,
                  json_file,
                  ensure_ascii=False,
                  indent=2)
    return file_path, ModifiedFileState.HAS_CREATED_NEW


def create_yaml_file(yaml_dict: dict,
                     file_path):
    # Enable PyYAML support for OrderedDict
    yaml.add_representer(
        OrderedDict,
        lambda dumper, data: dumper.represent_dict(
            getattr(data,
                    "viewitems" if sys.version_info < (3,) else "items")()),
        Dumper=yaml.SafeDumper)
    # Start dumping dict into yaml
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    with open(file_path, 'w', encoding='utf8') as yaml_file:
        yaml.safe_dump(yaml_dict,
                       yaml_file,
                       default_flow_style=False,
                       sort_keys=False)
    return file_path, ModifiedFileState.HAS_CREATED_NEW


def make_dirs(path, output_path):
    current_directory = os.getcwd() if output_path == '.' else output_path
    make_dir(current_directory, path)


def make_dir(current_directory, new_directory_name):
    directory = os.path.join(current_directory,
                             r'{}'.format(new_directory_name))
    try:
        os.makedirs(directory)
    except Exception as e:
        print(e)
        pass
    return directory


def validate_json_content(content_file: str):
    try:
        json.loads(content_file)
        return True
    except ValueError:
        return False


def validate_yaml_content(content_file: str):
    try:
        yaml.safe_load(content_file)
        return True
    except yaml.YAMLError:
        return False


def get_flutter_project_package():
    try:
        infile = open('pubspec.yaml', 'r')
        return infile.readline().replace('name:', '').strip()
    except (IOError, OSError):
        print('Please run this command in flutter folder.')
        exit(1)


def make_content_output_file(file_path: str, additional_info: str = None):
    sub_info = additional_info or ''
    return f'   └──{file_path:54}{sub_info:>34}'



from jinja2 import Environment, PackageLoader

from kntgen.file_helpers import *
from kntgen.str_helpers import auto_str


@auto_str
class Template(object):
    def __init__(self, folder_template: str):
        self.folder_template = folder_template
        self.env = Environment(
            loader=PackageLoader('kntgen_templates', folder_template),
            trim_blocks=True,
            lstrip_blocks=True
        )

    def create_file_from_template(self,
                                  class_name,
                                  file_extension='dart',
                                  template_file=None,
                                  sub_folder_template=None,
                                  include_root_folder=False,
                                  output_path=None,
                                  append_to_file=False,
                                  fixed_header: str = None,
                                  fixed_footer: str = None):
        if template_file is None:
            if file_extension:
                template_file = f'{class_name}.{file_extension}'
            else:
                template_file = class_name[len(self.name_lower):]

        env = self.env
        if sub_folder_template is not None:
            env = Environment(
                loader=PackageLoader(
                    'kntgen_templates',
                    f'{self.folder_template}/{sub_folder_template}'
                ),
                trim_blocks=True,
                lstrip_blocks=True
            )
        template = env.get_template(template_file)
        content = self.extract_content_from_template(template)

        if output_path is None:
            output_path = '../'

        if output_path.endswith('/'):
            output_path = output_path[:-1]

        if include_root_folder:
            folder = f'./{output_path}/{self.name_lower}'
        else:
            folder = f'./{output_path}'

        return create_file(
            content=content,
            file_name=class_name,
            file_extension=file_extension,
            folder=folder,
            append_to_file=append_to_file,
            fixed_header=fixed_header,
            fixed_footer=fixed_footer
        )

    def extract_content_from_template(self, template):
        return template.render()



from kntgen.file_helpers import *
from kntgen.model.flutter.flutter_raw_widget import *


class FWidget:
    MIN_OFFSET_PX = 4
    ALIGNMENT_OFFSET_RATIO = 0.14
    POSITIONED_STACK_OFFSET_RATIO = 0.16

    def __init__(self,
                 figma_node,
                 parent=None,
                 children: list = None):
        self.figma_node = figma_node
        self.parent = parent
        self.children = children or []
        # Extract info from node
        w_full_name, w_name, _ = FWidgetExtractor.regex_info(figma_node)
        self.w_raw_name = to_snake_case(
            w_full_name.replace(FWidgetExtractor.FW_CLICKABLE, '').replace(
                FWidgetExtractor.FW_DYNAMIC_DATA, '')
        )
        self.w_name = to_snake_case(w_name)
        self.need_data_object = FWidgetExtractor.FW_DYNAMIC_DATA in w_full_name
        self.clickable = FWidgetExtractor.FW_CLICKABLE in w_full_name
        self.rect = FWidgetExtractor.as_rect(self.figma_node)

    @property
    def w_id(self):
        return self.figma_node['id'] if 'id' in self.figma_node else ''

    @property
    def w_class(self):
        return 'SizedBox'

    @property
    def w_size(self):
        width = math.ceil(self.rect.width)
        height = math.ceil(self.rect.height)
        return width, height

    @property
    def w_size_present(self):
        width, height = self.w_size
        width_present = f'{width}.w'
        height_present = f'{height}.w'
        if FRawWidget.SCREEN_DESIGN_WIDTH - 5 < width:
            width_present = '1.sw'
        if FRawWidget.SCREEN_DESIGN_HEIGHT - 5 < height:
            height_present = '1.sh'
        return f'width: {width_present}', f'height: {height_present}'

    @property
    def w_on_tap_action_name(self):
        present_action = self.w_name if self.w_name else self.w_raw_name
        return f'_onTapped{to_pascal_case(present_action)}'

    @property
    def is_flex_size(self):
        """
        Whether this widget has flex size such as Text, List View... or not.
        This property will be used when we consider to wrap Expanded & Flexible in Column/Row's children

        :return:
        """
        return False

    @property
    def margin(self):
        """
        Calculate relative margin created between this widget and it's parent.
        In case widget's parent is Stack, then margin could be considered as a Positioned widget.

        :return: A ordered dictionary contains (left, top, right, bottom)
        """
        if not self.parent:
            return OrderedDict({
                Edge.LEFT: 0,
                Edge.TOP: 0,
                Edge.RIGHT: 0,
                Edge.BOTTOM: 0
            })
        parent_rect = self.parent.rect
        left_differ = max(0, math.ceil(self.rect.left - parent_rect.left))
        top_differ = max(0, math.ceil(self.rect.top - parent_rect.top))
        right_differ = max(0, math.ceil(parent_rect.right - self.rect.right))
        bottom_differ = max(0, math.ceil(parent_rect.bottom - self.rect.bottom))
        return OrderedDict({
            Edge.LEFT: left_differ,
            Edge.TOP: top_differ,
            Edge.RIGHT: right_differ,
            Edge.BOTTOM: bottom_differ
        })

    @property
    def padding(self):
        """
        Calculate distance from this widget to its children.

        :return: A ordered dictionary contains (left, top, right, bottom)
        """
        if not self.children:
            return OrderedDict({
                Edge.LEFT: 0,
                Edge.TOP: 0,
                Edge.RIGHT: 0,
                Edge.BOTTOM: 0
            })
        left_most = top_most = right_most = bottom_most = 0
        for c in self.children:
            rect = c.rect
            left_most = min(left_most, rect.left)
            top_most = min(top_most, rect.top)
            right_most = max(right_most, rect.right)
            bottom_most = max(bottom_most, rect.bottom)
        left_differ = max(0, math.ceil(left_most - self.rect.left))
        top_differ = max(0, math.ceil(top_most - self.rect.top))
        right_differ = max(0, math.ceil(self.rect.right - right_most))
        bottom_differ = max(0, math.ceil(self.rect.bottom - bottom_most))
        return OrderedDict({
            Edge.LEFT: left_differ,
            Edge.TOP: top_differ,
            Edge.RIGHT: right_differ,
            Edge.BOTTOM: bottom_differ
        })

    def generate(self,
                 strings_mapping: dict,
                 colors_mapping: dict,
                 images_mapping: dict):
        properties = ''.join([f'{p},' for p in
                              self.generate_properties(strings_mapping,
                                                       colors_mapping,
                                                       images_mapping) if p])
        properties_present = '' if not properties else f'\n{properties}'
        widget = f'{self.w_class}({properties_present})'
        # Check clickable
        if self.clickable:
            widget = f"""
RippleOverlay(
    onTap: {self.w_on_tap_action_name},
    child: {widget},
)
"""
        return self.generate_wrapper(widget)

    def generate_wrapper(self, generated_content):
        if not isinstance(self.parent, FViewGroup):
            return generated_content
        present_padding = FWidget.extract_edge_insets(
            self.margin,
            specific_edges=self.parent.needed_child_edge_insets
        )
        return f'Padding({present_padding},child:{generated_content},)' \
            if present_padding else generated_content

    def generate_child(self,
                       strings_mapping: dict,
                       colors_mapping: dict,
                       images_mapping: dict):
        if not self.children:
            return None
        total_child = len(self.children)
        prefix_children = 'child:' if total_child == 1 else 'children:['
        child_content = []
        for c in self.children:
            if not c:
                continue
            content = c.generate(
                strings_mapping=strings_mapping,
                colors_mapping=colors_mapping,
                images_mapping=images_mapping
            ) if not isinstance(c, FCustom) else c.instance_generate()
            content = self.generate_child_wrapper(c, content)
            if content:
                child_content.append(content)
        generate_children = ','.join(child_content)
        postfix_children = '' if total_child == 1 else ',]'
        return f'{prefix_children}{generate_children}{postfix_children}'

    def generate_child_wrapper(self, child_widget, generated_child_content):
        return generated_child_content

    def generate_properties(self,
                            strings_mapping: dict,
                            colors_mapping: dict,
                            images_mapping: dict):
        return [
            self.generate_child(strings_mapping=strings_mapping,
                                colors_mapping=colors_mapping,
                                images_mapping=images_mapping)
        ]

    @staticmethod
    def extract_dependencies_import(widget, imports: list):
        if imports is None:
            return
        for child in widget.children:
            if isinstance(child, FCustom):
                imports.append(child)
                continue
            FWidget.extract_dependencies_import(child, imports)

    @staticmethod
    def extract_interactive_functions(widget, functions: list):
        if functions is None:
            return
        if widget.clickable:
            functions.append(f'void {widget.w_on_tap_action_name}() {{}}')
        for child in widget.children:
            if isinstance(child, FCustom):
                continue
            FWidget.extract_interactive_functions(child, functions)

    @staticmethod
    def extract_edge_insets(edge_insets: OrderedDict,
                            specific_edges: list[Edge] = None,
                            inset_type: InsetType = InsetType.PADDING):
        """
        Present EdgeInsets

        :param edge_insets: A ordered dictionary contains insets at (left, top, right, bottom)
        :param specific_edges: Focus on which edge insets
        :param inset_type: Either EdgeInsets for padding or margin. Default is padding
        :return:
        """
        if not edge_insets or not specific_edges or len(edge_insets) < 4:
            return ''
        prefix = (inset_type or InsetType.PADDING).value
        named_constructor = 'only'
        edge_insets_present = ''
        # Fetch all edges that has an inset (> 0)
        edges = {e: edge_insets[e] for e in specific_edges if
                 edge_insets[e] > 0}
        vertical_symmetric = ''
        horizontal_symmetric = ''
        if Edge.TOP in edges and Edge.BOTTOM in edges:
            vertical_symmetric = f'vertical: {edges[Edge.TOP]}.w' \
                if edges[Edge.TOP] == edges[Edge.BOTTOM] else ''
        if Edge.LEFT in edges and Edge.RIGHT in edges:
            horizontal_symmetric = f'horizontal: {edges[Edge.LEFT]}.w' \
                if edges[Edge.LEFT] == edges[Edge.RIGHT] else ''

        # Consider appropriated EdgeInsets
        def _basic_present():
            return ''.join(
                [f'{e.value}:{v}.w,' for e, v in edges.items()])

        total_edges = len(edges)
        if total_edges == 1 or total_edges == 3:
            named_constructor = 'only'
            edge_insets_present = _basic_present()
        if total_edges == 2:
            if vertical_symmetric:
                named_constructor = 'symmetric'
                edge_insets_present = vertical_symmetric
            elif horizontal_symmetric:
                named_constructor = 'symmetric'
                edge_insets_present = horizontal_symmetric
            else:
                named_constructor = 'only'
                edge_insets_present = _basic_present()
        if total_edges == 4:
            if vertical_symmetric and horizontal_symmetric:
                named_constructor = 'symmetric'
                edge_insets_present = f'{vertical_symmetric},{horizontal_symmetric},'
            else:
                named_constructor = 'fromLTRB'
                edge_insets_present = ','.join(
                    [f'{v}.w' for v in edges.values()])
        return f'{prefix}:EdgeInsets.{named_constructor}({edge_insets_present})' \
            if edge_insets_present else ''

    @staticmethod
    def extract_image_path(node_id, images_mapping: dict):
        w_id_without_extension = (node_id or '').replace('+', '')
        if w_id_without_extension not in images_mapping:
            return None
        return images_mapping[w_id_without_extension].path

    @staticmethod
    def extract_colors(figma_node, colors_mapping: dict):
        if not figma_node:
            return None
        if 'fills' in figma_node and figma_node['fills']:
            first_bg = figma_node['fills'][0]
            if 'type' not in first_bg or not first_bg['type']:
                return None
            bg_type = first_bg['type']
            if bg_type == 'SOLID':
                return FWidgetExtractor.present_f_color(first_bg,
                                                        colors_mapping)
            if bg_type.startswith('GRADIENT'):
                # Guaranteed that each element in the below list is not empty
                gradient_hex_c_list = FWidgetExtractor.gradient_color_related(
                    first_bg)
                if not gradient_hex_c_list:
                    return None
                gradient_type = Gradient(bg_type) \
                    if Gradient.has_value(bg_type) \
                    else Gradient.LINEAR
                position_vector = [
                    Gradient.convert_coordinate_figma_to_flutter(coordinate) for
                    coordinate in gradient_hex_c_list[0][3]]
                first_pos_vector = f'{position_vector[0][0]}, {position_vector[0][1]}'
                second_pos_vector = f'{position_vector[1][0]}, {position_vector[1][1]}'
                gradient_width_vector = f'{math.fabs(position_vector[2][0])}'
                # Fetch named colors from mapping dict
                named_c_list = [FWidgetExtractor.present_f_color_from_scratch(
                    hex_color=hex_c[0],
                    alpha=hex_c[1],
                    colors_mapping=colors_mapping,
                    need_prefix=False
                ) for hex_c in gradient_hex_c_list]
                colors_present = ''.join(
                    [f'{named_c},' for named_c in named_c_list if named_c])
                stops = [f'{hex_c[2]:.2f}' for hex_c in gradient_hex_c_list]
                stops_present = ''.join(f'{s},\n' for s in stops)
                if gradient_type == Gradient.RADIAL:
                    position_vector_present = f'center: Alignment({first_pos_vector}),' \
                                              f'radius: {gradient_width_vector},'
                elif gradient_type == Gradient.SWEEP:
                    # TODO: Calculate [startAngle] & [endAngle] by device first_pos/second_pos
                    position_vector_present = f'center: Alignment({first_pos_vector}),' \
                                              'startAngle: 0.0,' \
                                              'endAngle: 360.0,'
                else:
                    position_vector_present = f'begin: Alignment({first_pos_vector}),' \
                                              f'end: Alignment({second_pos_vector}),'
                return f'gradient: {gradient_type.f_class_name()}(' \
                       f'{position_vector_present}' \
                       f'colors:[{colors_present}], ' \
                       f'stops:[{stops_present}],)'
        return None

    @staticmethod
    def extract_border(figma_node, colors_mapping: dict):
        if not figma_node:
            return None
        if 'strokes' not in figma_node:
            return None
        strokes = figma_node['strokes']
        if not strokes:
            return None
        stroke_weight = math.ceil(figma_node['strokeWeight']) \
            if 'strokeWeight' in figma_node else 1
        stroke_weight_present = '' if stroke_weight <= 1 else f'width: {stroke_weight}.w'
        solid_color_present = FWidgetExtractor.present_f_color(strokes[0],
                                                               colors_mapping)
        if not solid_color_present:
            # Determine if stroke color in gradient mode
            gradient_colors = FWidgetExtractor.gradient_color_related(
                strokes[0])
            if gradient_colors:
                # Just get the first color, we are currently not supporting gradient border
                first_stroke_color = FWidgetExtractor.hex_to_named_color(
                    hex_color=gradient_colors[0][0],
                    colors_mapping=colors_mapping
                )
                alpha_present = FWidgetExtractor.present_f_color_alpha(
                    gradient_colors[0][1])
                solid_color_present = f'color: {first_stroke_color}{alpha_present}'
        return f"border: Border.all({''.join([f'{p},' for p in [solid_color_present, stroke_weight_present] if p])})"

    @staticmethod
    def extract_border_radius(figma_node):
        if not figma_node:
            return None
        top_left = top_right = bottom_left = bottom_right = 0
        if 'cornerRadius' in figma_node:
            radius = math.ceil(figma_node['cornerRadius'])
            top_left = radius
            top_right = radius
            bottom_left = radius
            bottom_right = radius
        if 'rectangleCornerRadii' in figma_node:
            radii = figma_node['rectangleCornerRadii']
            if len(radii) == 4:
                top_left = math.ceil(radii[0])
                top_right = math.ceil(radii[1])
                bottom_left = math.ceil(radii[3])
                bottom_right = math.ceil(radii[2])
        if top_left > 0 and top_left == top_right \
                and top_right == bottom_right \
                and bottom_right == bottom_left:
            return f'borderRadius: BorderRadius.circular({top_left}.w)'
        border_radius_present = [
            '' if top_left <= 0 else f'topLeft: Radius.circular({top_left}.w)',
            '' if top_right <= 0 else f'topRight: Radius.circular({top_right}.w)',
            '' if bottom_left <= 0 else f'bottomLeft: Radius.circular({bottom_left}.w)',
            '' if bottom_right <= 0 else f'bottomRight: Radius.circular({bottom_right}.w)']
        border_radius_present_only = [r for r in border_radius_present if r]
        if not border_radius_present_only:
            return ''
        return f"borderRadius: BorderRadius.only({''.join([f'{r},' for r in border_radius_present_only])})"

    @staticmethod
    def extract_shadows(figma_node, colors_mapping: dict):
        if not figma_node:
            return None
        if 'effects' not in figma_node:
            return None
        if not figma_node['effects']:
            return None
        first_effect = figma_node['effects'][0]
        if 'type' not in first_effect or first_effect['type'] != 'DROP_SHADOW':
            return None
        if 'visible' not in first_effect or not first_effect['visible']:
            return None
        offset = first_effect['offset'] if 'offset' in first_effect else {}
        offset_x = math.ceil(offset['x']) if 'x' in offset else 0
        offset_y = math.ceil(offset['y']) if 'y' in offset else 0
        blur_radius = math.ceil(
            first_effect['radius']) if 'radius' in first_effect else 0
        # Present
        shadow_present = [FWidgetExtractor.present_f_color(first_effect,
                                                           colors_mapping),
                          '' if blur_radius <= 0 else f'blurRadius: {blur_radius}.w',
                          f'offset: Offset({offset_x}.w, {offset_y}.w)']
        return f"boxShadow: [BoxShadow({''.join([f'{s},' for s in shadow_present if s])}),]"


class FCustom(FWidget):
    def generate_build(self,
                       strings_mapping: dict,
                       colors_mapping: dict,
                       images_mapping: dict):
        first_view = self.children[0] if self.children else None
        return first_view.generate(
            strings_mapping=strings_mapping,
            colors_mapping=colors_mapping,
            images_mapping=images_mapping
        ) if first_view else 'const SizedBox()'

    def generate(self,
                 strings_mapping: dict,
                 colors_mapping: dict,
                 images_mapping: dict):
        imports = []
        functions = []
        FWidget.extract_dependencies_import(self, imports)
        FWidget.extract_interactive_functions(self, functions)
        imports_present = '\n'.join(
            [f'import {c.project_absolute_file_path()!r};' for c in imports
             if c])
        functions_present = '\n\n'.join(functions)
        return f"""
import 'package:flutter/material.dart';
import 'package:knt_ui/knt_ui.dart';
import '{self.package_name}/resources.dart';
{imports_present}

class {self.w_class} extends StatelessWidget {{
    const {self.w_class}({{Key? key}}) : super(key: key);
    
    {functions_present}
    
    @override
    Widget build(BuildContext context) {{
        return {self.generate_build(strings_mapping=strings_mapping,
                                    colors_mapping=colors_mapping,
                                    images_mapping=images_mapping)};
    }}
}}
"""

    def instance_generate(self):
        return f'const {self.w_class}()'

    def generated_file_name(self):
        return self.w_raw_name

    def generate_container_folder(self):
        if FWidgetExtractor.FW_CUSTOM_ITEM in self.w_raw_name:
            return 'items'
        return 'custom'

    def project_absolute_file_path(self):
        return f'{self.package_name}/ui/{self.generate_relative_file_path()}.dart'

    def generate_relative_file_path(self):
        return f'{self.generate_container_folder()}/{self.generated_file_name()}'

    @property
    def w_class(self):
        return to_pascal_case(self.w_raw_name)

    @property
    def package_name(self):
        project_package = get_flutter_project_package()
        return f'package:{project_package}'


class FViewGroup(FWidget):
    def generate_child(self,
                       strings_mapping: dict,
                       colors_mapping: dict,
                       images_mapping: dict):
        if not self.children:
            return None
        child_content = []
        total_child = len(self.children)
        for i in range(total_child):
            child = self.children[i]
            if not child:
                continue
            content = child.generate(
                strings_mapping=strings_mapping,
                colors_mapping=colors_mapping,
                images_mapping=images_mapping
            ) if not isinstance(child, FCustom) else child.instance_generate()
            content = self.generate_child_wrapper(child, content)
            if content:
                child_content.append(content)
            if self.main_alignment != Alignment.SPACE_BETWEEN and i < total_child - 1:
                if self.direction == Direction.VERTICAL:
                    distance_to_next_child = math.ceil(
                        self.children[i + 1].rect.top - child.rect.bottom
                    )
                    if distance_to_next_child > 0:
                        child_content.append(
                            f'SizedBox(height: {distance_to_next_child}.w)')
                if self.direction == Direction.HORIZONTAL:
                    distance_to_next_child = math.ceil(
                        self.children[i + 1].rect.left - child.rect.right
                    )
                    if distance_to_next_child > 0:
                        child_content.append(
                            f'SizedBox(width: {distance_to_next_child}.w)')
        generate_children = ','.join(child_content)
        return f'children: [{generate_children},]'

    def generate_child_wrapper(self, child_widget, generated_child_content):
        if not generated_child_content:
            return generated_child_content
        # If child widget is a view group fixed style, then we need provide size wrapper for it
        # By default, we use its presented child's size as parent size
        if not isinstance(child_widget, FViewGroupFixedStyle):
            return generated_child_content
        # If parent view group has only 1 child view group fixed style then we just wrap it with Expanded
        if self.total_view_group_fixed_style_child == 1:
            return f'Expanded(child:{generated_child_content},)'
        # Otherwise wrap by Flexible with flex proportion
        present_size = child_widget.presented_child_size
        if not present_size:
            return generated_child_content
        if self.direction == Direction.HORIZONTAL:
            # Provide height when parent view group is a row
            flex_proportion = f'{present_size[0]} / {self.w_size[0]}'
        else:
            # Provide width when parent view group is a column
            flex_proportion = f'{present_size[1]} / {self.w_size[1]}'
        return f'Flexible(flex: {flex_proportion},' \
               f'child:{generated_child_content},)'

    def generate_properties(self,
                            strings_mapping: dict,
                            colors_mapping: dict,
                            images_mapping: dict):
        properties = [self.generate_direction()]
        properties.extend(self.generate_alignments())
        properties.append(self.generate_child(
            strings_mapping=strings_mapping,
            colors_mapping=colors_mapping,
            images_mapping=images_mapping
        ))
        return properties

    def generate_alignments(self) -> list[str]:
        present_main_size = 'mainAxisSize: MainAxisSize.min' \
            if self.total_view_group_fixed_style_child > 0 else ''
        present_main_alignment = f'mainAxisAlignment: MainAxisAlignment' \
                                 f'.{self.main_alignment.value}'
        present_cross_alignment = f'crossAxisAlignment: CrossAxisAlignment' \
                                  f'.{self.cross_alignment.value}'
        if self.main_alignment == Alignment.START:
            present_main_alignment = ''
        if self.cross_alignment == Alignment.CENTER:
            present_cross_alignment = ''
        return [
            present_main_size,
            present_main_alignment,
            present_cross_alignment,
        ]

    def generate_direction(self) -> str:
        return f"direction: Axis.{'horizontal' if self.direction == Direction.HORIZONTAL else 'vertical'}"

    @property
    def direction(self):
        return Direction.VERTICAL

    @property
    def x_alignment(self):
        w_center_x = self.rect.center_x
        alignment_offset = self.rect.width * FWidget.ALIGNMENT_OFFSET_RATIO
        # View group aligns as a column
        if self.direction == Direction.VERTICAL:
            for c_rect in self.children_rect:
                center_x = c_rect.center_x
                if center_x < w_center_x - alignment_offset:
                    return Alignment.START
                if center_x > w_center_x + alignment_offset:
                    return Alignment.END
            # Last check: left edge and right edge of each child
            left_edges_count = 1
            right_edges_count = 1
            first_c_left = self.children_rect[0].left
            first_c_right = self.children_rect[0].right
            total_child = len(self.children_rect)
            for i in range(1, total_child):
                c_rect = self.children_rect[i]
                if first_c_left - FWidget.MIN_OFFSET_PX < c_rect.left \
                        < first_c_left + FWidget.MIN_OFFSET_PX:
                    left_edges_count += 1
                if first_c_right - FWidget.MIN_OFFSET_PX < c_rect.right \
                        < first_c_right + FWidget.MIN_OFFSET_PX:
                    right_edges_count += 1
            if left_edges_count == total_child:
                return Alignment.START
            if right_edges_count == total_child:
                return Alignment.END
            return Alignment.CENTER
        # Otherwise view group aligns as a row
        left_children_edge = self.children[0].rect.left
        right_children_edge = self.children[-1].rect.right
        if left_children_edge > self.rect.left + alignment_offset \
                and right_children_edge < self.rect.right - alignment_offset:
            return Alignment.CENTER
        if left_children_edge > self.rect.left + alignment_offset:
            return Alignment.END
        if right_children_edge < self.rect.right - alignment_offset:
            return Alignment.START
        if len(self.children) == 2:
            left_margin = self.children[1].rect.left - self.children[
                0].rect.right
            if left_margin >= alignment_offset:
                return Alignment.SPACE_BETWEEN
        return Alignment.START

    @property
    def y_alignment(self):
        w_center_y = self.rect.center_y
        alignment_offset = self.rect.height * FWidget.ALIGNMENT_OFFSET_RATIO
        # View group aligns as a row
        if self.direction == Direction.HORIZONTAL:
            for c_rect in self.children_rect:
                center_y = c_rect.center_y
                if center_y < w_center_y - alignment_offset:
                    return Alignment.START
                if center_y > w_center_y + alignment_offset:
                    return Alignment.END
            # Last check: top edge and bottom edge of each child
            top_edges_count = 1
            bottom_edges_count = 1
            first_c_top = self.children_rect[0].top
            first_c_bottom = self.children_rect[0].bottom
            total_child = len(self.children_rect)
            for i in range(1, len(self.children_rect)):
                c_rect = self.children_rect[i]
                if first_c_top - FWidget.MIN_OFFSET_PX < c_rect.top \
                        < first_c_top + FWidget.MIN_OFFSET_PX:
                    top_edges_count += 1
                if first_c_bottom - FWidget.MIN_OFFSET_PX < c_rect.bottom \
                        < first_c_bottom + FWidget.MIN_OFFSET_PX:
                    bottom_edges_count += 1
            if top_edges_count == total_child:
                return Alignment.START
            if bottom_edges_count == total_child:
                return Alignment.END
            return Alignment.CENTER
        # Otherwise view group aligns as a column
        top_children_edge = self.children[0].rect.top
        bottom_children_edge = self.children[-1].rect.bottom
        if top_children_edge > self.rect.top + alignment_offset \
                and bottom_children_edge < self.rect.bottom - alignment_offset:
            return Alignment.CENTER
        if top_children_edge > self.rect.top + alignment_offset:
            return Alignment.END
        if bottom_children_edge < self.rect.bottom - alignment_offset:
            return Alignment.START
        if len(self.children) == 2:
            top_margin = self.children[1].rect.top - self.children[
                0].rect.bottom
            if top_margin >= alignment_offset:
                return Alignment.SPACE_BETWEEN
        return Alignment.START

    @property
    def children_rect(self):
        return [c.rect for c in self.children if c]

    @property
    def main_alignment(self):
        return Alignment.START

    @property
    def cross_alignment(self):
        return Alignment.CENTER

    @property
    def needed_child_edge_insets(self):
        specific_edges = []
        if self.direction == Direction.VERTICAL:
            if self.cross_alignment == Alignment.START:
                specific_edges.append(Edge.LEFT)
            if self.cross_alignment == Alignment.END:
                specific_edges.append(Edge.RIGHT)
        if self.direction == Direction.HORIZONTAL:
            if self.cross_alignment == Alignment.START:
                specific_edges.append(Edge.TOP)
            if self.cross_alignment == Alignment.END:
                specific_edges.append(Edge.BOTTOM)
        return specific_edges

    @property
    def total_view_group_fixed_style_child(self):
        total = 0
        for child in self.children:
            if isinstance(child, FViewGroupFixedStyle):
                total += 1
        return total


class FViewGroupFixedStyle(FViewGroup):
    def __init__(self, figma_node, children: list = None, is_horizontal=False):
        super().__init__(figma_node, children)
        self.is_horizontal = is_horizontal

    def generate_properties(self, strings_mapping: dict, colors_mapping: dict,
                            images_mapping: dict):
        properties = super().generate_properties(strings_mapping,
                                                 colors_mapping,
                                                 images_mapping)
        properties.insert(0, self.generate_scroll_physic())
        return properties

    def generate_child(self,
                       strings_mapping: dict,
                       colors_mapping: dict,
                       images_mapping: dict):
        first_child = 'const SizedBox()'
        if self.children:
            child = self.children[0]
            if child:
                first_child = child.generate(
                    strings_mapping=strings_mapping,
                    colors_mapping=colors_mapping,
                    images_mapping=images_mapping
                ) if not isinstance(child,
                                    FCustom) else child.instance_generate()
                first_child = self.generate_child_wrapper(child, first_child)
        item_builder = f'itemBuilder: (context, index) => {first_child}'
        item_count = 'itemCount: 1'
        return ','.join([item_builder, item_count])

    def generate_direction(self) -> str:
        return 'scrollDirection: Axis.horizontal' \
            if self.direction == Direction.HORIZONTAL else None

    def generate_scroll_physic(self):
        return 'physics: const BouncingScrollPhysics()'

    def generate_alignments(self) -> list[str]:
        return []

    @property
    def direction(self):
        return Direction.HORIZONTAL if self.is_horizontal \
            else Direction.VERTICAL

    @property
    def presented_child(self):
        return self.children[0] if self.children else None

    @property
    def presented_child_size(self):
        return self.presented_child.w_size if self.presented_child else None



from os.path import splitext
from urllib.parse import urlparse

from sheetfu import SpreadsheetApp

from kntgen.file_helpers import *
from kntgen.model.figma_models import Color
from kntgen.model.flutter.flutter_raw_widget import FWidgetExtractor
from kntgen.model.ui_config import *
from kntgen.parallel_helpers import run_parallel_tasks
from kntgen.printer import *
from kntgen.str_helpers import *


class FDefined:
    NAMED_COLOR_ENDPOINT = 'https://api.color.pizza/v1/'
    GOOGLE_FONT_ENDPOINT = 'https://www.googleapis.com/webfonts/v1/webfonts'
    ASSETS_FLUTTER_GEN = 'Assets.images.'
    ASSETS_FLUTTER_PATH = '.path'
    COLOR_FLUTTER_GEN = 'ColorName.'
    LOCALE_KEYS = 'LocaleKeys.'
    LOCALE_KEYS_EXTENSION = '.tr()'


class NamedLocaleText:
    @staticmethod
    def localized(custom_widget_name: str, text_name: str):
        prefix, name_without_type = FWidgetExtractor.localized_key_from_name(
            custom_widget_name)
        text_name_without_type = FWidgetExtractor.name_without_type(text_name)
        if not name_without_type or not text_name_without_type:
            return None
        text_name_with_page = f'{to_camel_case(prefix)}_' \
                              f'{to_camel_case(name_without_type)}_' \
                              f'{to_camel_case(text_name_without_type)}'
        return f'{FDefined.LOCALE_KEYS}' \
               f'{text_name_with_page}' \
               f'{FDefined.LOCALE_KEYS_EXTENSION}'


class NamedColor:
    def __init__(self, name_duplicated_times=1, **kwargs):
        self.name = kwargs['name'] if 'name' in kwargs else ''
        self.hex = kwargs['hex'] if 'hex' in kwargs else ''
        self.requested_hex = kwargs[
            'requestedHex'] if 'requestedHex' in kwargs else ''
        self.name_duplicated_times = name_duplicated_times or 0
        # Handle name with format
        snake_case_name = to_snake_case(
            ''.join([c for c in self.name if 'a' <= c.lower() <= 'z']))
        postfix = '' if self.name_duplicated_times <= 0 \
            else f'_{(self.name_duplicated_times + 1)}'
        self.snake_case_latin_name = f'{snake_case_name}{postfix}'
        self.camel_case_latin_name = to_camel_case(self.snake_case_latin_name)

    @property
    def flutter_gen_color(self):
        return f'{FDefined.COLOR_FLUTTER_GEN}{self.camel_case_latin_name}'


class NamedImage:
    def __init__(self, image_name: str, image_format: str):
        resource_name = (image_name or '') \
            .replace(FWidgetExtractor.FW_CLICKABLE, '') \
            .replace(FWidgetExtractor.FW_DYNAMIC_DATA, '')
        if resource_name.startswith(FWidgetExtractor.FW_CUSTOM_WIDGET):
            resource_name = resource_name.replace(
                FWidgetExtractor.FW_CUSTOM_WIDGET, '')
        self.image_name = ''.join(
            [c for c in (resource_name or '') if not c.isspace()])
        self.path = f'{FDefined.ASSETS_FLUTTER_GEN}' \
                    f'{to_camel_case(self.image_name)}' \
                    f'{FDefined.ASSETS_FLUTTER_PATH}'
        self.image_format = image_format or 'svg'


class ImageResource:
    def __init__(self, url: str,
                 folder: str,
                 image_name: str,
                 format_image: str):
        self.url = url
        self.folder = folder or DEFAULT_PATH_ASSETS_IMAGES
        self.image_name = image_name
        self.format_image = format_image or 'png'


class FontResource:
    def __init__(self, font_url, folder, file_name):
        self.font_url = font_url
        self.folder = folder or DEFAULT_PATH_ASSETS_FONTS
        self.file_name = file_name


def fetch_named_colors(hex_color_list: list):
    """
    Fetch name of input colors.
    Duplicated name with differ colors will add an extra number postfix .e.g pearl_1

    :param hex_color_list: List colors in RGA hex format
    :return: Dictionary of named color with (k, v) = (hex_color, dict_color_from_api)
    """
    if not hex_color_list:
        return []
    hex_as_id_colors = ','.join(
        [hex_color[1:] for hex_color in hex_color_list])
    response = requests.get(
        f'{FDefined.NAMED_COLOR_ENDPOINT}?values={hex_as_id_colors}'
    )
    named_colors = {}
    if response.status_code == 200:
        res_dict = json.loads(response.text)
        if 'colors' in res_dict:
            named_list = []
            for color in res_dict['colors']:
                if 'hex' not in color \
                        or 'name' not in color \
                        or 'requestedHex' not in color:
                    continue
                c_name = color['name']
                name_duplicated_times = named_list.count(c_name)
                named_color = NamedColor(
                    name_duplicated_times=name_duplicated_times,
                    **color
                )
                named_colors[named_color.requested_hex] = named_color
                named_list.append(c_name)
    return named_colors


def download_images(figma,
                    file_id,
                    folder: str,
                    images: dict,
                    format_image='png',
                    scale_image=2):
    downloaded_file_paths = []
    download_info_images = figma.get_file_images(
        file_key=file_id,
        ids=images.keys(),
        format=format_image,
        scale=scale_image
    )
    if not download_info_images or not download_info_images.images:
        return downloaded_file_paths
    # Prepare resources
    image_sources: list[ImageResource] = []
    for image_id, url in download_info_images.images.items():
        node = {}
        if image_id in images.keys():
            node = images[image_id]
        if not node:
            continue
        node_name = node['name'] if 'name' in node else ''
        image_name = to_snake_case(node_name)
        image_sources.append(
            ImageResource(
                url=url,
                folder=folder,
                image_name=image_name,
                format_image=format_image))
    # Start downloading
    downloaded_file_paths.extend(
        run_parallel_tasks(download_single_image, image_sources))
    return downloaded_file_paths


def download_single_image(resource: ImageResource):
    file_ext = get_extension(resource.url)
    if not file_ext:
        file_ext = f'.{resource.format_image}'
    # Download and save image into file
    file_path = f'{resource.folder}/{resource.image_name}{file_ext}'
    downloaded_file_path = download_to_file(resource.url, file_path)
    return make_content_output_file(
        file_path=downloaded_file_path,
        additional_info=ModifiedFileState.HAS_CREATED_NEW
    )


def download_fonts(google_font_dict: dict, api_key: str, folder: str):
    downloaded_file_paths = []
    google_fonts_url = f'{FDefined.GOOGLE_FONT_ENDPOINT}?key={api_key}'
    gf_response = requests.get(google_fonts_url)
    if gf_response.status_code >= 300:
        error_printer('Could not download fonts!')
        return downloaded_file_paths
    gf_fonts_dict = json.loads(gf_response.text)
    gf_fonts = gf_fonts_dict['items'] if 'items' in gf_fonts_dict else []
    checked_font_family = 0
    # Prepare resources
    font_sources: list[FontResource] = []
    for font in gf_fonts:
        if 'family' not in font and 'files' not in font:
            continue
        family_font = font['family']
        font_files = font['files']

        # Get styles of this needed font
        if family_font in google_font_dict:
            for style in google_font_dict[family_font]:
                if style[0] in font_files:
                    font_url = font_files[style[0]]
                    file_name = style[1]
                    font_sources.append(
                        FontResource(
                            font_url=font_url,
                            folder=folder,
                            file_name=file_name))
            # Update progress
            checked_font_family += 1
            if checked_font_family == len(google_font_dict):
                break
    # Start downloading
    downloaded_file_paths.extend(
        run_parallel_tasks(download_single_font, font_sources))
    return downloaded_file_paths


def download_single_font(resource: FontResource):
    file_ext = get_extension(resource.font_url)
    if not file_ext:
        file_ext = '.ttf'
    # Download and save image into file
    file_path = f'{resource.folder}/{resource.file_name}{file_ext}'
    downloaded_file_path = download_to_file(resource.font_url, file_path)
    return make_content_output_file(
        file_path=downloaded_file_path,
        additional_info=ModifiedFileState.HAS_CREATED_NEW
    )


def beauty_print_named_colors(named_colors: dict[str:NamedColor]):
    content_colors = []
    for hex_c, color in (named_colors or {}).items():
        color_tuple = Color.hex_to_rgb(hex_c)
        colored_content = colored_printer(
            *color_tuple,
            text=f'    {color.name:44}{hex_c:>8}    '
        )
        content_colors.append(colored_content)
    if len(content_colors) > 0:
        beauty_printer(
            title=f'{len(content_colors)} colors in design file',
            content=content_colors,
            max_char_inline=60
        )


def create_google_spreadsheet(design_text_dict, package_name):
    sa = SpreadsheetApp(
        './../kntgen_templates/ui/secret/figma_text_generator_gs.json')
    # email_editor = input('Enter your email as an editor of Google Spreadsheet: ')
    email_editor = 'kiennt.ftx@gmail.com'
    spreadsheet = sa.create(
        name='Appixi',
        editor=email_editor
    )
    spreadsheet.create_sheets([package_name])
    project_sheet = spreadsheet.get_sheet_by_name(package_name)
    # Create columns
    title_columns_range = project_sheet.get_range_from_a1('A1:C1')
    title_columns_range.set_values([
        ['localized_key', 'screen', 'en']
    ])
    # Fill rows
    total_design_text = 0
    for k, v in design_text_dict.items():
        total_design_text += len(v)
    data_range = project_sheet.get_range_from_a1(
        f'A2:C{total_design_text + 1}')
    data = []
    for page, texts_in_page in design_text_dict.items():
        for content_text in texts_in_page:
            data.append(['key', page, content_text])
    data_range.set_values(data)


def get_extension(url):
    """Return the filename extension from url, or ''."""
    parsed = urlparse(url)
    root, ext = splitext(parsed.path)
    return ext



from os import walk

from googletrans import Translator
from jinja2 import Environment, PackageLoader

from kntgen.collection_helpers import *
from kntgen.ui.ui_external_source import *


class FGenerator:
    DEFAULT_LOCALE_FILE_EXTENSION = 'json'
    REGEX_DETECT_LOCALE = '^(?P<languageCode>[a-z]{2})(?P<countryCode>-[A-Z]{2})?.json$'
    MAP_PATH_DELIMITER = '.'

    @staticmethod
    def generate_json_locale_file(text_dict: OrderedDict = None,
                                  locale_folder_path=None,
                                  pivot_locale=None,
                                  need_print=True):
        if not text_dict:
            return
        generated_content = []
        # Guarantee non null variables
        nn_locale_folder_path = locale_folder_path \
                                or DEFAULT_PATH_ASSETS_TRANSLATIONS
        nn_pivot_locale = pivot_locale or DEFAULT_LOCALE
        # Create new locale keys
        new_string_dict = OrderedDict()
        for page, text_in_page in text_dict.items():
            prefix, page_without_type = FWidgetExtractor.localized_key_from_name(
                page)
            if not page_without_type:
                continue
            page_without_type = to_camel_case(page_without_type)
            for text_id, text_name_with_content in text_in_page.items():
                text_id_without_type = FWidgetExtractor.name_without_type(
                    text_name_with_content[0])
                if not text_id_without_type:
                    continue
                text_id_without_type = to_camel_case(text_id_without_type)
                if prefix not in new_string_dict:
                    new_string_dict[prefix] = OrderedDict()
                if page_without_type not in new_string_dict[prefix]:
                    new_string_dict[prefix][page_without_type] = OrderedDict()
                new_string_dict[prefix][page_without_type][
                    text_id_without_type] = text_name_with_content[1]
        # Check if locale file has content, then collect all
        file_path = f'{nn_locale_folder_path}/{nn_pivot_locale}.{FGenerator.DEFAULT_LOCALE_FILE_EXTENSION}'
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, 'r', encoding='utf-8') as pivot_locale_file:
            pivot_locale_dict = json.loads(pivot_locale_file.read(),
                                           object_pairs_hook=OrderedDict)
        if pivot_locale_dict:
            # Merge with existed keys
            ordered_dict_merge(result_d=pivot_locale_dict,
                               complement_d=new_string_dict)
            file_path, _ = create_json_file(pivot_locale_dict,
                                            file_path)
            additional_info = ModifiedFileState.HAS_APPENDED
        else:
            file_path, additional_info = create_json_file(new_string_dict,
                                                          file_path)
        generated_content.append(
            make_content_output_file(file_path, additional_info))

        if need_print:
            beauty_printer(
                title='Successfully generated localization file',
                content=generated_content
            )

    @staticmethod
    def generate_color_file(named_colors: dict[str: NamedColor] = None,
                            folder: str = None,
                            need_print=True):
        if not named_colors:
            return
        header_file = """<?xml version="1.0" encoding="utf-8"?>
<resources>
"""
        footer_file = '</resources>'
        body_file = ''
        for hex_named_color, color in named_colors.items():
            if not color:
                continue
            name_snake_case = color.snake_case_latin_name
            body_file = body_file + f'    <color name="{name_snake_case}">{hex_named_color}</color>\n'
        file_path, additional_info = create_file(
            content=f'{header_file}{body_file}{footer_file}',
            file_name='colors',
            file_extension='xml',
            folder=folder or DEFAULT_PATH_ASSETS_COLORS
        )
        if need_print:
            beauty_printer(
                title='Successfully generated colors file',
                content=[make_content_output_file(file_path, additional_info)]
            )

    @staticmethod
    def generate_ui_file(f_gen_list: dict[str: str],
                         need_print=True):
        generated_files = []
        for path, f_w in (f_gen_list or {}).items():
            if not path or not f_w:
                continue
            path_component = path.split('/')
            folder = '/'.join(path_component[:-1])
            file_name = path_component[-1]
            file_path, additional_info = create_file(
                content=f_w,
                file_name=to_snake_case(file_name),
                file_extension='dart',
                folder=f'./lib/ui/{folder}'
            )
            generated_files.append(
                make_content_output_file(file_path, additional_info))
        # Also generate [resources.dart] which exports all generated files
        # by Flutter [build_runner]
        resource_path, modified_info = FGenerator._generate_ui_file(
            template_file_name='resources.dart',
            file_name='resources',
            package_name=get_flutter_project_package()
        )
        generated_files.append(
            make_content_output_file(resource_path, modified_info))
        if need_print and generated_files:
            beauty_printer(
                title='Successfully generated widget file',
                content=generated_files
            )

    @staticmethod
    def translate_locales(locale_folder_path=None,
                          pivot_locale=None,
                          supported_locale: list = None,
                          placeholder=None,
                          need_print=True):
        nn_locale_folder_path = locale_folder_path \
                                or DEFAULT_PATH_ASSETS_TRANSLATIONS
        nn_pivot_locale = pivot_locale or DEFAULT_LOCALE
        nn_supported_locale = supported_locale or []
        with open(
                f'{nn_locale_folder_path}/{nn_pivot_locale}.{FGenerator.DEFAULT_LOCALE_FILE_EXTENSION}',
                'r',
                encoding='utf-8'
        ) as pivot_locale_file:
            pivot_locale_dict = json.loads(pivot_locale_file.read(),
                                           object_pairs_hook=OrderedDict)
        # Translate lacked locale keys from other locales when compare to pivot locale
        # pivot_locale_dict = new_string_dict
        if not nn_supported_locale:
            # In case user doesn't define supported locales, then we browse
            # through locale folder to file all locales as much as possible
            _, _, filenames = next(walk(nn_locale_folder_path),
                                   (None, None, []))
            for filename in filenames:
                match_filename_info = re.search(FGenerator.REGEX_DETECT_LOCALE,
                                                filename)
                if match_filename_info:
                    language_code = match_filename_info.group('languageCode')
                    # Current [googletrans] library just support [languageCode]
                    locale = language_code
                    if locale != pivot_locale and locale not in nn_supported_locale:
                        nn_supported_locale.append(locale)
        if not nn_supported_locale or not pivot_locale:
            error_printer('Could not find any locale to translate!')
            return
        generated_content = []
        if pivot_locale_dict and nn_supported_locale:
            generated_files = run_parallel_tasks(
                FGenerator.generate_translated_locale,
                nn_supported_locale,
                nn_pivot_locale,
                nn_locale_folder_path,
                pivot_locale_dict,
                placeholder)
            generated_content.extend(generated_files)

        if need_print:
            beauty_printer(
                title='Successfully translated localization file',
                content=generated_content
            )

    @staticmethod
    def generate_translated_locale(locale: str,
                                   pivot_locale: str,
                                   folder_path: str,
                                   pivot_locale_dict: dict = None,
                                   placeholder: str = None):
        if not locale:
            error_printer('Locale is not valid!')
            return
        if not pivot_locale:
            error_printer('Source locale is not valid!')
            return
        if not folder_path:
            error_printer('Could not find translations folder!')
            return
        if not pivot_locale_dict:
            error_printer('Source of translation is not defined!')
            return

        # Start translating
        info_printer(f'Translating all text into locale [{locale}] ...',
                     emphasis=False)
        with open(
                f'{folder_path}/{locale}.{FGenerator.DEFAULT_LOCALE_FILE_EXTENSION}',
                'r',
                encoding='utf-8'
        ) as locale_file:
            locale_dict = json.loads(locale_file.read(),
                                     object_pairs_hook=OrderedDict)
        # Compare with pivot locale and find lacked keys, then create a complement dict
        complement_dict = OrderedDict()
        FGenerator._find_complement_of_2_dicts(
            complement_dict=complement_dict,
            pivot_dict=pivot_locale_dict,
            lacked_dict=locale_dict
        )
        # Translate
        locale_map_path = OrderedDict()
        FGenerator._convert_locale_tree_to_map_path(
            locale_dict=complement_dict,
            map_path=locale_map_path
        )
        keys_to_translate = [k for k in locale_map_path.keys()]
        strings_to_translate = [v for v in locale_map_path.values()]
        translator = Translator()
        translations = []
        # [googletrans] library currently has an issue about bulk mode
        for s in strings_to_translate:
            try:
                translation = translator.translate(
                    s,
                    dest=f'{locale}-cn' if locale == 'zh' else locale,
                    src=pivot_locale
                )
                print(f'🌏 {translation.origin} -> {translation.text}')
                translations.append(translation)
            except TypeError:
                translations.append(
                    f'{placeholder or DEFAULT_PLACEHOLDER_TRANSLATE} [{s}]')
        if len(translations) != len(keys_to_translate):
            error_printer('Oops... [googletrans] mismatch error!')
            return
        for i in range(len(translations)):
            t = translations[i]
            locale_map_path[keys_to_translate[i]] = t \
                if isinstance(t, str) else t.text
        complement_dict.clear()
        FGenerator._convert_map_path_to_locale_tree(
            map_path=locale_map_path,
            locale_dict=complement_dict
        )
        # Merge complement dict into full locale dict
        full_locale_dict = OrderedDict()
        FGenerator._merge_complement_to_lacked(
            complement_dict=complement_dict,
            lacked_dict=locale_dict,
            pivot_dict=pivot_locale_dict,
            result_dict=full_locale_dict
        )
        # Print lacked locale files which filled with complement locale keys
        file_path = f'{folder_path}_complement/{locale}.{FGenerator.DEFAULT_LOCALE_FILE_EXTENSION}'
        return make_content_output_file(*create_json_file(full_locale_dict,
                                                          file_path))

    @staticmethod
    def _generate_ui_file(template_file_name: str,
                          file_name: str,
                          folder: str = None,
                          *render_args,
                          **render_kwargs):
        env = Environment(
            loader=PackageLoader('kntgen_templates', 'ui'),
            trim_blocks=True,
            lstrip_blocks=True
        )
        template = env.get_template(template_file_name)
        content = template.render(*render_args, **render_kwargs)
        return create_file(
            content=content,
            file_name=file_name,
            file_extension='dart',
            folder=f'./lib/{folder}' if folder else './lib')

    @staticmethod
    def _find_complement_of_2_dicts(complement_dict: OrderedDict,
                                    pivot_dict: OrderedDict,
                                    lacked_dict: OrderedDict):
        if complement_dict is None or pivot_dict is None or lacked_dict is None:
            return
        for k, v in pivot_dict.items():
            if k not in lacked_dict:
                complement_dict[k] = v
            else:
                if not isinstance(v, OrderedDict):
                    continue
                if k not in complement_dict:
                    complement_dict[k] = OrderedDict()
                FGenerator._find_complement_of_2_dicts(
                    complement_dict=complement_dict[k],
                    pivot_dict=v,
                    lacked_dict=lacked_dict[k]
                )
                # Remove sub dict if it's empty
                if not complement_dict[k]:
                    complement_dict.pop(k, None)

    @staticmethod
    def _merge_complement_to_lacked(complement_dict: OrderedDict,
                                    lacked_dict: OrderedDict,
                                    pivot_dict: OrderedDict,
                                    result_dict: OrderedDict,
                                    previous_key=None):
        if complement_dict is None or pivot_dict is None \
                or lacked_dict is None or result_dict is None:
            return
        for k, v in pivot_dict.items():
            if k in lacked_dict:
                result_dict[k] = lacked_dict[k]
                previous_key = k
                if not isinstance(v, OrderedDict):
                    continue
                # Check dry-run if still has locale key need to be complemented in this lacked sub-dict
                if k in complement_dict:
                    FGenerator._merge_complement_to_lacked(
                        complement_dict=complement_dict[k],
                        lacked_dict=lacked_dict[k],
                        pivot_dict=v,
                        result_dict=result_dict[k]
                    )
                    previous_key = None
            else:
                insert_at_pos_ordered_dict(
                    dic=result_dict,
                    current_key_dict=k,
                    item=complement_dict[k],
                    key_to_insert_after=previous_key
                )
                previous_key = k

    @staticmethod
    def _convert_locale_tree_to_map_path(locale_dict: OrderedDict,
                                         map_path: OrderedDict,
                                         last_key: str = None):
        def _append_path(path, new_path):
            return f'{path}{FGenerator.MAP_PATH_DELIMITER}{new_path}' \
                if path else new_path

        def _remove_last_path(path):
            components = (path or '').split(FGenerator.MAP_PATH_DELIMITER)
            if not components:
                return ''
            components.pop()
            return FGenerator.MAP_PATH_DELIMITER.join(components)

        if locale_dict is None or map_path is None:
            return
        for k, v in locale_dict.items():
            if not isinstance(v, str) and not isinstance(v, OrderedDict):
                continue
            last_key = _append_path(last_key, k)
            if isinstance(v, str):
                map_path[last_key] = v
            if isinstance(v, OrderedDict):
                FGenerator._convert_locale_tree_to_map_path(
                    locale_dict=v,
                    map_path=map_path,
                    last_key=last_key
                )
            last_key = _remove_last_path(last_key)

    @staticmethod
    def _convert_map_path_to_locale_tree(map_path: OrderedDict,
                                         locale_dict: OrderedDict):
        if map_path is None or locale_dict is None:
            return
        for path, text in map_path.items():
            temp_dict = locale_dict
            components = (path or '').split(FGenerator.MAP_PATH_DELIMITER)
            if not components:
                continue
            for component in components[:-1]:
                if component not in temp_dict:
                    temp_dict[component] = OrderedDict()
                temp_dict = temp_dict[component]
            temp_dict[components[-1]] = text



# coding=utf-8
from kntgen.cmd import *
from kntgen.file_helpers import *
from kntgen.model.json import JSON
from kntgen.printer import beauty_printer
from kntgen.str_helpers import to_snake_case


class JsonCommand(Command):
    def __init__(self, model_name, options):
        super(JsonCommand, self).__init__()
        self.model_name = model_name
        # For future usage
        self.options = CommandOption(options)

    def create_files(self,
                     json_str: str,
                     folder: str = None,
                     need_beauty_print=True):
        json_template = JSON(
            model_name=self.model_name,
            json_str=json_str
        )
        content_files = json_template.create_model_files()
        file_path, additional_info = create_file(
            content=content_files,
            file_name=to_snake_case(self.model_name),
            file_extension='dart',
            folder=folder or './lib/model'
        )
        if need_beauty_print:
            beauty_printer(
                title='Successfully created model file',
                content=[make_content_output_file(file_path,
                                                  additional_info)]
            )
        return file_path, additional_info



import json

import requests

from kntgen.model.figma_models import *


class FigmaPy:
    def __init__(self, token, oauth2=False):
        self.api_uri = 'https://api.figma.com/v1/'
        self.token_uri = 'https://www.figma.com/oauth'
        self.api_token = token
        self.oauth2 = oauth2

    # -------------------------------------------------------------------------
    # FIGMA API
    # -------------------------------------------------------------------------
    '''
    Request Figma API
    '''

    def api_request(self, endpoint, method='get', payload=None):
        method = method.lower()

        if payload is None:
            payload = ''

        if self.oauth2:
            header = {'Authorization': 'Bearer {0}'.format(self.api_token)}
        else:
            header = {'X-Figma-Token': '{0}'.format(self.api_token)}

        header['Content-Type'] = 'application/json'

        try:
            if method == 'head':
                response = requests.head(
                    '{0}{1}'.format(self.api_uri, endpoint), headers=header)
            elif method == 'delete':
                response = requests.delete(
                    '{0}{1}'.format(self.api_uri, endpoint), headers=header)
            elif method == 'get':
                response = requests.get('{0}{1}'.format(self.api_uri, endpoint),
                                        headers=header, data=payload)
            elif method == 'options':
                response = requests.options(
                    '{0}{1}'.format(self.api_uri, endpoint), headers=header)
            elif method == 'post':
                response = requests.post(
                    '{0}{1}'.format(self.api_uri, endpoint), headers=header,
                    data=payload)
            elif method == 'put':
                response = requests.put('{0}{1}'.format(self.api_uri, endpoint),
                                        headers=header, data=payload)
            else:
                response = None
            if response.status_code == 200:
                return json.loads(response.text)
            else:
                return None
        except (
                Exception,
                requests.HTTPError,
                requests.exceptions.SSLError
        ) as e:
            print(f'Error occurred attempting to make an api request. {str(e)}')
            return None

    # -------------------------------------------------------------------------
    # OAUTH2
    # -------------------------------------------------------------------------
    """
    Create token from client_id, client_secret, code
    """

    def create_token(self, client_id, client_secret, redirect_uri, code):
        payload = {
            'client_id': '{0}'.format(client_id),
            'client_secret': '{0}'.format(client_secret),
            'grant_type': 'authorization_code',
            'redirect_uri': '{0}'.format(redirect_uri),
            'code': '{0}'.format(code)
        }
        try:
            response = requests.post(self.token_uri, data=payload)
            print(response.text)
            if response.status_code == 200:
                token_data = json.loads(response.text)
                return [token_data['access_token'], token_data['expires_in']]
            else:
                return None
        except requests.HTTPError:
            print('HTTP Error occurred while trying to generate access token.')
            return None

    """
    Returns API user's access token
    """

    def get_access_token(self):
        return self.api_token

    # -------------------------------------------------------------------------
    # SCOPE: FILES
    # -------------------------------------------------------------------------
    """
    Get the JSON file contents for a file.
    """

    def get_file(self, file_key, geometry=None, version=None):
        optional_data = ''
        if geometry is not None or version is not None:
            optional_data = '?'
            if geometry is not None:
                optional_data += str(geometry)
                if version is not None:
                    optional_data += '&version={0}'.format(str(version))
            elif version is not None:
                optional_data += str(version)

        data = self.api_request('files/{0}{1}'.format(file_key, optional_data),
                                method='get')
        if data is not None:
            return File(data['name'], data['document'], data['components'],
                        data['lastModified'], data['thumbnailUrl'],
                        data['schemaVersion'], data['styles'])

    def get_file_nodes(self, file_key, ids=None):
        """
        Returns the nodes referenced to by :ids as a JSON object.
        The nodes are retrieved from the Figma file referenced to by :file_key.
        """
        id_array = []
        for id in ids:
            id_array.append(id)
        id_list = ','.join(id_array)
        data = self.api_request(
            'files/{0}/nodes/?ids={1}'.format(file_key, id_list),
            method='get')
        if data is not None:
            return FileNode(data['name'], data['version'],
                            data['lastModified'], data['thumbnailUrl'],
                            data['role'], data['nodes'])

    """
    Get the version history of a file.
    """

    def get_file_versions(self, file_key):
        data = self.api_request('files/{0}/versions'.format(file_key),
                                method='get')
        if data is not None:
            return FileVersions(data['versions'], data['pagination'])

    """
    Get all comments on a file.
    """

    def get_comments(self, file_key):
        data = self.api_request('files/{0}/comments'.format(file_key),
                                method='get')
        if data is not None:
            return Comments(data['comments'])

    """
    Create a comment on a file.
    """

    def post_comment(self, file_key, message, client_meta=None):
        print(client_meta)
        if client_meta is not None:
            payload = '{{"message":"{0}","client_meta":{1}}}'.format(
                message.title(), client_meta)
        else:
            payload = "{{'message':'{0}'}}".format(message)
        data = self.api_request('files/{0}/comments'.format(file_key),
                                method='post', payload=payload)
        if data is not None:
            return Comment(data['id'], data['file_key'], data['parent_id'],
                           data['user'], data['created_at'],
                           data['resolved_at'], data['message'],
                           data['client_meta'], data['order_id'])

    # -------------------------------------------------------------------------
    # SCOPE: IMAGES
    # -------------------------------------------------------------------------
    """
    Get urls for server-side rendered images from a file.
    """

    def get_file_images(self, file_key, ids, scale=None, format=None,
                        version=None):
        optional_data = ''
        if scale is not None or format is not None or version is not None:
            if scale is not None:
                optional_data += '&scale={0}'.format(str(scale))
            if format is not None:
                optional_data += '&format={0}'.format(str(format))
            if version is not None:
                optional_data += '&version={0}'.format(str(version))
        id_array = []
        for id in ids:
            id_array.append(id)
        id_list = ','.join(id_array)
        data = self.api_request(
            'images/{0}?ids={1}{2}'.format(file_key, id_list, optional_data),
            method='get')
        if data is not None:
            return FileImages(data['images'], data['err'])

    # -------------------------------------------------------------------------
    # SCOPE: TEAMS
    # -------------------------------------------------------------------------
    """
    Get all projects for a team
    """

    def get_team_projects(self, team_id):
        data = self.api_request('teams/{0}/projects'.format(team_id),
                                method='get')
        if data is not None:
            return TeamProjects(data['projects'])

    # -------------------------------------------------------------------------
    # SCOPE: PROJECTS
    # -------------------------------------------------------------------------
    """
    Get all files for a project
    """

    def get_project_files(self, project_id):
        data = self.api_request('projects/{0}/files'.format(project_id))
        if data is not None:
            return ProjectFiles(data['files'])
